/*
  Program..: IMSERVER20.PRG
  Author...: Roger J. Donnay
  Date.....: June 11, 2015
*/

#include "SET.CH"        // alaska
#include "dbstruct.ch"   // alaska
#include "dmlb.ch"       // alaska
#include "common.ch"     // alaska
#include "class.ch"      // alaska
#include "error.ch"      // alaska
#include "fileio.ch"     // alaska
#include "dcdialog.ch"   // eXpress++
#include "dcicon.ch"     // eXpress++
#include "dchtml.ch"     // eXpress++
#include "appevent.ch"   // alaska
#include "service.ch"    // alaska
#INCLUDE "directry.CH"   // alaska
#INCLUDE "dll.CH"        // alaska

#pragma Library("asinet10.lib")  // needed for sending email
#pragma Library("asiutl10.lib")  // needed for sending email

#define CERT_FILE         'IMCERT.PEM'
#define APP_NAME          'Instant Messenger Server'
#define APP_DESCRIPTION   'Instant Messenger Server'
#define DEF_PORT          '80'
#define ERR_LOG_FILE      'ERROR.LOG'

#define USERID_LEN         50
#define PASSWORD_LEN       20

#DEFINE IP_ADDRESS_HOST   'LocalHost'

FIELD USER_ID, DATE, TIME, FROM_USER, TO_USER, LAST_TO, NET_LOGIN, GROUP_NAME, ;
      CONV_NMBR, MSG_READ

FIELD mail_to, mail_from, error_code, date_sent

#xtranslate NTrim(<n>)   =>  LTrim(Str(<n>))

STATIC snRefreshInterval := 0, soRefreshTimer, snTimeOut := 0, scParamLine := '', ;
       scRootDir, scHostAlias, scStatus := '', snMode := 0, slActive := .t., ;
       saPreferences := {}, scUserGroup := '', snDays, scUserId, slIsRunning := .t., ;
       slIsPaused := .f., soServer, saClientList := {}, soBrowse, saGetList

STATIC saUsers[0], slRefreshBrowse
STATIC scServerVersion := '2.0.1'

* -----------

PROC AppSys
RETURN

* -----------

PROCEDURE Main( cParams )

IF Empty(cParams)
  WebService():start()
ELSE
  ServerMain(,cParams)
ENDIF

RETURN

* -------------

CLASS WebService From ServiceApp
EXPORTED:
  CLASS METHOD main
  CLASS METHOD stop
  CLASS METHOD pause
  CLASS METHOD continue
  CLASS VAR server
  CLASS VAR mainWindow
HIDDEN:
  CLASS VAR isRunning
ENDCLASS

* -------------

CLASS METHOD WebService:main()

ServerMain(self)

RETURN self

* -------------

CLASS METHOD WebService:stop()
::isRunning := .F.
slIsRunning := .f.
slIsPaused := .f.
Sleep(5)

IIF(Valtype(::server)=='O',::server:destroy(),nil)
LogActivity( nil, 'Server Stopped' )
PostAppEvent(xbeP_Close,,,::mainWindow)

RETURN self

* -------------

CLASS METHOD WebService:pause()
IIF(Valtype(::server)=='O',::server:stop(),nil)
slIsRunning := .f.
slIsPaused := .t.
LogActivity( nil, 'Server Paused' )
RETURN self

* -------------

CLASS METHOD WebService:continue()
IIF(Valtype(::server)=='O',::server:start(),nil)
slIsRunning := .t.
slIsPaused := .f.
LogActivity( nil, 'Server Started' )
RETURN self

* ============

CLASS Session

EXPORTED:

VAR sessionId
VAR seconds
VAR lastTime
VAR lastDate

VAR userId
VAR password

* -----------

INLINE METHOD Init( cSessionID, lRenew )

DEFAULT lReNew := .F.
IF !lRenew
  ::sessionID   := cSessionID
  ::seconds     := Seconds()
  ::lastTime    := Time()
  ::lastDate    := Date()
  ::userID      := ''
  ::password    := ''
ENDIF

RETURN self

ENDCLASS

* =============

STATIC PROCEDURE ServerMain( oService, cParams )

   LOCAL GetList[0], GetOptions, lStarted := .t., oBrowse
   LOCAL oThread, lRunning
   LOCAL nPort, cDomain, cTitle
   LOCAL aData := { 80, ;
                    Pad(IP_ADDRESS_HOST,100), ;
                    'HostAlias' }

   LOCAL aRef := { 'Port', 'Domain', 'HostAlias' }
   LOCAL cIniFile := DC_Path(AppName(.t.)) + 'IMSERVER.INI'
   LOCAL aBrowPres

   DC_ChDir( DC_Path(AppName(.t.)))
   CreateDirectoryA('Users')
   CreateDirectoryA('Photos')
   CreateDirectoryA('VoiceClips')

   saGetList := getList

   DC_AutoRestoreWindow( { HKEY_LOCAL_MACHINE, 'Software\Donnay Software\IMServer\Windows' } )

   DC_IniLoad( cIniFile, 'IMSERVER', aData, aRef )
   nPort := aData[1]
   cDomain := Alltrim(aData[2])
   scHostAlias := Alltrim(aData[3])
   slRefreshBrowse := .t.
   aData := { 300, 10 }
   aRef := { 'TimeOut','Days'}
   DC_IniLoad( cIniFile, 'MESSAGES', aData, aRef )
   snTimeOut := aData[1]
   snDays := aData[2]

   DC_IconDefault(ICON_EXPRESS)
   _LoadRdds()
   UpdateUserFile( cIniFile )

   OpenUserFile(.t.) // update structure if needed
   OpenGroupFile(.t.) // update structure if needed
   OpenMessageFile(.t.) // update structure if needed
   IMGROUPS->(dbCloseArea())
   IMUSERS->(dbCloseArea())
   IMMSG->(dbCloseArea())

   cTitle := 'Instant Messenger Web/SOAP Service ( Host IP: ' + cDomain + ;
       ' Port: ' + Alltrim(Str(nPort))

   IF oService == NIL .AND. DC_IsAppRunning('XbpDialog',cTitle,,.t.)
     QUIT
   ENDIF

   OpenLogFile()

   soServer := HttpEndpoint():new( nPort )

   Thread():new():start({||SendMailThread(@lRunning,GetList)})

   aBrowPres := ;
    { { XBP_PP_COL_HA_FGCLR, GRA_CLR_WHITE },            /* Header FG Color  */  ;
      { XBP_PP_COL_HA_BGCLR, GRA_CLR_DARKGRAY },         /* Header BG Color  */  ;
      { XBP_PP_COL_DA_ROWSEPARATOR, XBPCOL_SEP_DOTTED }, /* Row Sep          */  ;
      { XBP_PP_COL_DA_COLSEPARATOR, XBPCOL_SEP_DOTTED }, /* Col Sep          */  ;
      { XBP_PP_COL_DA_FGCLR, GRA_CLR_BLACK },            /* Row FG Color     */  ;
      { XBP_PP_COL_DA_BGCLR, GRA_CLR_WHITE },            /* Row BG Color     */  ;
      { XBP_PP_COL_DA_ROWHEIGHT, -1 },                   /* Row Height       */  ;
      { XBP_PP_COL_HA_HEIGHT, 10 },                      /* Header Height    */  ;
      { XBP_PP_HILITE_FGCLR, GRA_CLR_WHITE },            /* Hilite FG color  */  ;
      { XBP_PP_HILITE_BGCLR, GRA_CLR_BLUE },             /* Hilite BG color  */  ;
      { XBP_PP_COL_FA_FGCLR, GRA_CLR_WHITE },            /* Footer FG Color  */  ;
      { XBP_PP_COL_FA_BGCLR, GRA_CLR_DARKGRAY },         /* Footer BG Color  */  ;
      { XBP_PP_COL_FA_HEIGHT, 10 }                       /* Footer Height    */  ;
    }

   @ 1,0 DCPUSHBUTTON CAPTION 'Exit' SIZE 11,1.2 ;
         ACTION {||DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)}

   @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION {||IIF(lStarted,'Stop','Start')} SIZE 11,1.2 ;
         ACTION {||IIF(!lStarted, soServer:start(), nil ),;
                   IIF(lStarted, (slIsPaused := .t.,soServer:stop()), nil ),;
                   lStarted := !lStarted, ;
                   LogActivity(oBrowse,Str(ThreadID(),3) + IIF(lStarted," Server started"," Server stopped")), ;
                   DC_GetRefresh(GetList,,,,'BUTTONS')} ;
         GROUP 'BUTTONS'

   @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Users' SIZE 11,1.2 ;
         ACTION {||Thread():new():start({||UserMaint()})} ;
         TOOLTIP 'Maintain the User Database' ;
         GROUP 'BUTTONS'

   @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Groups' SIZE 11,1.2 ;
         ACTION {||Thread():new():start({||GroupMaint()})} ;
         TOOLTIP 'Maintain the Group Database' ;
         GROUP 'BUTTONS'

   @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Messages' SIZE 11,1.2 ;
         ACTION {||Thread():new():start({||MessageMaint()})} ;
         TOOLTIP 'View/Maintain the Message Database' ;
         GROUP 'BUTTONS'

   @ DCGUI_ROW, DCGUI_COL + 10 DCPUSHBUTTON CAPTION 'Clients' SIZE 11,1.2 ;
         WHEN {||lStarted} ;
         ACTION {||ViewClients()}

   @ DCGUI_ROW, DCGUI_COL + 10 DCCHECKBOX slRefreshBrowse PROMPT 'Refresh Browse on each hit'

   @ 2.5, 0 DCSAY 'Logged on Users:' SAYSIZE 0 ;
      HYPERLINK {||IIF(Empty(saUsers),DC_WinAlert('No Users'),ShowUsers())} ;

   @ DCGUI_ROW, DCGUI_COL + 10 DCSAY {||LoggedOnUsers()} ;
      SAYSIZE 10 SAYLEFT ;
      COLOR GRA_CLR_RED FONT '12.Lucida Console Bold' ID 'USERS'

   @ DCGUI_ROW, DCGUI_COL + 20 DCPUSHBUTTON CAPTION 'Refresh User Count' SIZE 15 ;
      ACTION {||DC_GetRefresh(GetList,'USERS')}

   @ DCGUI_ROW, DCGUI_COL + 20 DCSAY 'eXpress++ IM Server (' + scServerVersion + ')' SAYSIZE 0 ;
      SAYBOTTOM FONT '10.Courier New'

  @ 3.5,0 DCBROWSE oBrowse SIZE 100,20 FONT '8.Courier New' ;
         PRESENTATION aBrowPres ALIAS 'IMSERVER' ;
         NOHSCROLL ;
         COLOR {||IIF(IMSERVER->(Recno())%2=0,{nil,GRA_CLR_WHITE},{nil,GRA_CLR_CYAN})}

   DCBROWSECOL FIELD IMSERVER->date HEADER 'Date' WIDTH 10 PARENT oBrowse
   DCBROWSECOL FIELD IMSERVER->time HEADER 'Time' WIDTH 8 PARENT oBrowse
   DCBROWSECOL DATA {||IIF(IMSERVER->ssl,'SSL','')} HEADER 'SSL' WIDTH 3 PARENT oBrowse
   DCBROWSECOL FIELD IMSERVER->info HEADER 'Info' WIDTH 80 PARENT oBrowse

   lStarted := .t.

   IF oService == NIL

     DCGETOPTIONS ;
       ABORTQUERY ;
       EXITQUERY MSG {||_QuitApp(oBrowse)} ;
       CLOSEQUERY MSG {||_QuitApp(oBrowse)} ;
       AUTORESIZE TABSTOP

   ELSE

     DCGETOPTIONS ;
       AUTORESIZE TABSTOP

   ENDIF

   IF Valtype(oService) = 'O'
     oService:server := soServer
   ENDIF

   DCREAD GUI FIT ;
     TITLE cTitle ;
     OPTIONS GetOptions SETAPPWINDOW ;
     EVAL {|o|IIF(Valtype(soServer)=='O',(soServer:start(),LogActivity(oBrowse,'Server Started')),nil), ;
              IIF(Valtype(oService)=='O',oService:mainWindow := o,nil), ;
              soBrowse := oBrowse }

   // don't forget to destroy the object in order to release socket handles!
   IF Valtype(oService) = 'O'
     oService:stop()
   ENDIF
   soServer:stop()

Return

* ------------

STATIC FUNCTION LoggedOnUsers()

LOCAL aUsers[0], i, oClient, nClients, cUserId

nClients := Len(saClientList)

FOR i := 1 TO nClients
  oClient := saClientList[i]:requestClient
  IF Valtype(oClient) == 'O'
    cUserID := oClient:getVar("USERID")
    IF AScan( aUsers, {|c|c == Upper(Trim(cUserID))} ) == 0
      AAdd( aUsers, Upper(Trim(cUserId)) )
    ENDIF
  ENDIF
NEXT

RETURN Len(aUsers)

* ------------

STATIC FUNCTION Cleanup( GetList, oServer )

LOCAL i, nFound, aClients, lRefresh := .f.

DO WHILE .T.
  Sleep(500)
  aClients := saClientList
  i := 1
  DO WHILE i <= Len(saUsers)
    nFound := AScan( aClients, {|o|DC_XtoC(o:cargo) == saUsers[i,1]})
    IF nFound == 0 .OR. (Seconds() - saUsers[nFound,3]) > snTimeOut
      ARemove(saUsers,i)
      lRefresh := .T.
      EXIT
    ELSEIF nFound <= Len(saUsers) .AND. (Seconds() - saUsers[nFound,3]) > snTimeOut
      ARemove(saUsers,i)
      lRefresh := .T.
      EXIT
    ENDIF
    i++
  ENDDO
  IF lRefresh
    UserFlag('$$$',200)
    DC_GetRefresh(saGetList,'USERS')
  ENDIF
  lRefresh := .F.
ENDDO

RETURN nil

* ------------

STATIC FUNCTION UserMaint()

STATIC slIsRunning := .f.

LOCAL GetList[0], GetOptions, oTab1, oTab2, oBrowse, oToolBar1, oToolBar2, ;
      lDeleted := Set(_SET_DELETED,.t.), oGet1, oAppWindow := SetAppWindow(), ;
      oUsers, lAddMode := .f., lEditMode := .f., nSaveRec := 0

IF slIsRunning
  RETURN nil
ENDIF

slIsRunning := .t.

OpenUserFile()

oUsers := IMUSERS->(DC_DbRecord():new())
IMUSERS->(DC_DbScatter(oUsers))

@ 0,0 DCTABPAGE oTab1 SIZE 100,23 CAPTION 'Browse' ;
      GOTFOCUS {||oBrowse:refreshAll()} ;
      WHEN {||!lEditMode}

@ 0,0 DCTABPAGE oTab2 RELATIVE oTab1 CAPTION 'Detail' ;
      GOTFOCUS {||DC_GetRefresh(GetList,,,,'TAB2'), ;
                  SetAppFocus(oGet1), ;
                  IMUSERS->(dbRLock())}

@ 2,2 DCBROWSE oBrowse ALIAS 'IMUSERS' SIZE 96,20 ;
      PRESENTATION DC_BrowPres() ;
      PARENT oTab1 ;
      WHEN {||!lEditMode} ;
      ITEMMARKED {||IMUSERS->(DC_DbScatter(oUsers))} ;
      ITEMSELECTED {||DC_TabActivate(oTab2), ;
                      DC_GetRefresh(GetList,,,,'TAB2')} ;
      COLOR {||IIF(AScan(saUsers,{|a|a[1]==Alltrim(IMUSERS->user_id)})>0, ;
                  {nil,GRA_CLR_CYAN},{nil,GRA_CLR_WHITE})}

DCBROWSECOL FIELD IMUSERS->user_id HEADER 'User ID' WIDTH 10 PARENT oBrowse
DCBROWSECOL FIELD IMUSERS->user_last HEADER 'Last Name' WIDTH 12 PARENT oBrowse
DCBROWSECOL FIELD IMUSERS->user_first HEADER 'First Name' WIDTH 10 PARENT oBrowse
DCBROWSECOL FIELD IMUSERS->user_alias HEADER 'Alias' WIDTH 8 PARENT oBrowse
DCBROWSECOL FIELD IMUSERS->status HEADER 'Status' WIDTH 10 PARENT oBrowse
DCBROWSECOL FIELD IMUSERS->active HEADER 'Active' WIDTH 4 PARENT oBrowse

@ 23,0 DCTOOLBAR oToolBar1 SIZE 100 BUTTONSIZE 18,1
@ 24,0 DCTOOLBAR oToolBar2 SIZE 100 BUTTONSIZE 18,1

* ------- tool bar 1 ---------

DCADDBUTTON CAPTION 'New User' PARENT oToolBar1 ;
   ACTION {||nSaveRec := IMUSERS->(RecNo()), ;
             IMUSERS->(dbGoTo(0)), ;
             IMUSERS->(DC_DbScatter(oUsers)), ;
             lAddMode := .t., ;
             lEditMode := .t., ;
             DC_TabActivate(oTab2)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Edit User' PARENT oToolBar1 ;
   ACTION {||nSaveRec := IMUSERS->(RecNo()), ;
             lEditMode := .t., ;
             DC_GetRefresh(GetList), ;
             DC_TabActivate(oTab2)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Save' PARENT oToolBar1 ;
   ACTION {||IMUSERS->(DC_DbGather(oUsers,lAddMode)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList)} ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Cancel' PARENT oToolBar1 ;
   ACTION {||IMUSERS->(dbGoTo(nSaveRec)), ;
             IMUSERS->(DC_DbScatter(oUsers)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList) } ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Delete User' PARENT oToolBar1 ;
   ACTION {||IIF(DC_MsgBox(,,{'Delete this user?'},,,,.t.), ;
             (IMUSERS->(DC_RecLock()), ;
              IMUSERS->(DC_Blank()), ;
              IMUSERS->(DC_RecLock()), ;
              IMUSERS->(dbDelete()), ;
              IMUSERS->(dbRUnlock()), ;
              IMUSERS->(dbSkip(-1))),nil), ;
             DC_GetRefresh(GetList)} ;
   WHEN {||!lEditMode}


* ------ tool bar 2 -------

DCADDBUTTON CAPTION 'Logoff this User' PARENT oToolBar2 ;
   ACTION {||IIF(DC_MsgBox(,,{'Logoff this user?'},,,,.t.), ;
             (UserFlag(Alltrim(IMUSERS->user_id),303), ;
              Sleep(200), ;
              UserFlag(Alltrim(IMUSERS->user_id),303)),nil), ;
             DC_GetRefresh(GetList)}

DCADDBUTTON CAPTION 'Logoff ALL Users' PARENT oToolBar2 ;
   ACTION {||IIF(DC_MsgBox(,,{'Logoff ALL users?'},,,,.t.), ;
             (UserFlag('$$$',303), ;
              Sleep(200), ;
              UserFlag('$$$',303)),nil), ;
             DC_GetRefresh(GetList)}

DCADDBUTTON CAPTION 'ShutDown ALL Users' PARENT oToolBar2 ;
   ACTION {||IIF(DC_MsgBox(,,{'Shut Down ALL users?'},,,,.t.), ;
             (UserFlag('$$$',304), ;
              Sleep(200), ;
              UserFlag('$$$',304), ;
              MemoWrit('..\CLIENT\SHUTDOWN.TXT', ;
                      'Shutdown requested by IM Server Administrator')),nil), ;
              DC_GetRefresh(GetList)}

DCADDBUTTON CAPTION 'Enable ALL Users' PARENT oToolBar2 ;
   ACTION {||IIF(DC_MsgBox(,,{'Enable ALL users?'},,,,.t.), ;
             Ferase('..\CLIENT\SHUTDOWN.TXT'),nil)}

DCADDBUTTON CAPTION 'Refresh Browse' PARENT oToolBar2 ;
   ACTION {||oBrowse:refreshAll()}

DCSETPARENT TO oTab2
DCSETGROUP TO 'TAB2'

@ 2,2 DCSAY 'User ID' GET oUsers:user_id PICTURE '@!' PARENT oTab2 GETOBJECT oGet1
@ 3,2 DCSAY 'Last Name' GET oUsers:user_last PARENT oTab2
@ 4,2 DCSAY 'First Name' GET oUsers:user_first PARENT oTab2
@ 5,2 DCSAY 'Alias' GET oUsers:user_alias PARENT oTab2
@ 6,2 DCSAY 'Password' PARENT oTab2
@ DCGUI_ROW, DCGUI_COL + 10 DCGET oUsers:password PARENT oTab2
@ 7,2 DCSAY 'Email' GET oUsers:user_email PARENT oTab2
@ 8,2 DCSAY 'Phone' GET oUsers:user_phone PARENT oTab2
@ 9,2 DCSAY 'Group(s)' GET oUsers:user_group PARENT oTab2 GETSIZE 50 ;
      POPUP {|c|PopGroups(c)}
@10,2 DCSAY 'Status' GET oUsers:status PARENT oTab2
@11,2 DCSAY 'Language' GET oUsers:language PARENT oTab2
@12,2 DCSAY 'Client Version' GET oUsers:client_ver PARENT oTab2

@14,2 DCCHECKBOX oUsers:active PROMPT 'This user is Active' PARENT oTab2

DCGETOPTIONS SAYRIGHTBOTTOM SAYWIDTH 100 EDITPROTECT {||!lEditMode} TABSTOP
DCREAD GUI FIT OPTIONS GetOptions TITLE 'User Maintenance' SETAPPWINDOW

CLOSE ALL
Set(_SET_DELETED,lDeleted)

SetAppWindow(oAppWindow)

slIsRunning := .f.

RETURN nil

* ------------

STATIC FUNCTION MessageMaint()

STATIC slIsRunning := .f.

LOCAL GetList[0], GetOptions, oTab1, oTab2, oBrowse, oToolBar, ;
      lDeleted := Set(_SET_DELETED,.t.), oGet1, oAppWindow := SetAppWindow(), ;
      oMsg, lAddMode := .f., lEditMode := .f., nSaveRec := 0

IF slIsRunning
  RETURN nil
ENDIF

slIsRunning := .t.

OpenMessageFile()
IMMSG->(OrdSetFocus('DATE'))

oMsg := IMMSG->(DC_DbRecord():new())
IMMSG->(DC_DbScatter(oMsg))

@ 0,0 DCTABPAGE oTab1 SIZE 100,23 CAPTION 'Browse' ;
      GOTFOCUS {||oBrowse:refreshAll()} ;
      WHEN {||!lEditMode}

@ 0,0 DCTABPAGE oTab2 RELATIVE oTab1 CAPTION 'Detail' ;
      GOTFOCUS {||DC_GetRefresh(GetList,,,,'TAB2'), ;
                  SetAppFocus(oGet1), ;
                  IMMSG->(dbRLock())}

@ 2,2 DCBROWSE oBrowse ALIAS 'IMMSG' SIZE 96,15 ;
      PRESENTATION DC_BrowPres() ;
      FONT '8.Lucida Console' ;
      PARENT oTab1 ;
      WHEN {||!lEditMode} ;
      ITEMMARKED {||IMMSG->(DC_DbScatter(oMsg)),DC_GetRefresh(GetList,'MESSAGE')} ;
      ITEMSELECTED {||DC_TabActivate(oTab2), ;
                      DC_GetRefresh(GetList,,,,'TAB2')} ;
      COLOR {||IIF(AScan(saUsers,{|a|a[1]==Alltrim(IMMSG->to_user)})>0, ;
                  {nil,GRA_CLR_CYAN},{nil,GRA_CLR_WHITE})}

DCBROWSECOL FIELD IMMSG->date HEADER 'Date' WIDTH 10 PARENT oBrowse ;
   SORT {||IMMSG->(OrdSetFocus('DATE'))} DEFAULT

DCBROWSECOL FIELD IMMSG->time HEADER 'Time' WIDTH 8 PARENT oBrowse ;
   SORT {||IMMSG->(OrdSetFocus('DATE'))}

DCBROWSECOL FIELD IMMSG->conv_nmbr HEADER 'Conv #' WIDTH 10 PICTURE '9999999999' PARENT oBrowse ;
   SORT {||IMMSG->(OrdSetFocus('CONV_NMBR'))}

DCBROWSECOL FIELD IMMSG->to_user HEADER 'To User' WIDTH 20 PARENT oBrowse ;
   SORT {||IMMSG->(OrdSetFocus('TO_USER'))}

DCBROWSECOL FIELD IMMSG->from_user HEADER 'From User' WIDTH 20 PARENT oBrowse ;
   SORT {||IMMSG->(OrdSetFocus('TO_USER'))}

DCBROWSECOL DATA {||IIF(IMMSG->msg_read,'Yes','No')} HEADER 'Read' WIDTH 10 PARENT oBrowse

@17.5, 2 DCMULTILINE oMsg:message SIZE 95,5 NOHSCROLL FONT '10.Lucida Console' PARENT oTab1 ;
      ID 'MESSAGE'

@ 23,0 DCTOOLBAR oToolBar SIZE 100 BUTTONSIZE 12,1

DCADDBUTTON CAPTION 'New Msg' PARENT oToolBar ;
   ACTION {||nSaveRec := IMMSG->(RecNo()), ;
             IMMSG->(dbGoTo(0)), ;
             IMMSG->(DC_DbScatter(oMsg)), ;
             lAddMode := .t., ;
             lEditMode := .t., ;
             DC_TabActivate(oTab2)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Edit Msg' PARENT oToolBar ;
   ACTION {||nSaveRec := IMMSG->(RecNo()), ;
             lEditMode := .t., ;
             DC_GetRefresh(GetList), ;
             DC_TabActivate(oTab2)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Save' PARENT oToolBar ;
   ACTION {||IMMSG->(DC_DbGather(oMsg,lAddMode)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList)} ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Cancel' PARENT oToolBar ;
   ACTION {||IMMSG->(dbGoTo(nSaveRec)), ;
             IMMSG->(DC_DbScatter(oMsg)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList) } ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Delete Msg' PARENT oToolBar ;
   ACTION {||IIF(DC_MsgBox(,,{'Delete this message?'},,,,.t.), ;
             (IMMSG->(DC_RecLock()), ;
              IMMSG->(DC_Blank()), ;
              IMMSG->(dbSkip(-1))),nil), ;
             DC_GetRefresh(GetList)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Refresh' PARENT oToolBar ;
   ACTION {||oBrowse:refreshAll()} ;
   WHEN {||!lEditMode}

DCSETPARENT TO oTab2
DCSETGROUP TO 'TAB2'

@ 2,2 DCSAY 'Conv #' GET oMsg:conv_nmbr PICTURE '9999999999'
@ 3,2 DCSAY 'To User' GET oMsg:to_user PICTURE '@!' PARENT oTab2 GETOBJECT oGet1
@ 4,2 DCSAY 'From User' GET oMsg:from_user PARENT oTab2
@ 5,2 DCSAY 'Date' GET oMsg:date PARENT oTab2 POPUP {|d|DC_PopDate(d,,,,,,2)}
@ 6,2 DCSAY 'Time' GET oMsg:time PARENT oTab2
@ 7,2 DCSAY 'Type' GET oMsg:msg_type PICT '9' PARENT oTab2
@ 8,2 DCSAY 'Wave' GET oMsg:wave PARENT oTab2
@ 9,2 DCSAY 'FG Color' GET oMsg:fgcolor PICT '999' PARENT oTab2
@10,2 DCSAY 'BG Color' GET oMsg:bgColor PICT '999' PARENT oTab2
@11,2 DCSAY 'Font' GET oMsg:font PARENT oTab2 GETSIZE 50
@12,2 DCSAY 'File Name' GET oMsg:file_name PARENT oTab2 GETSIZE 80
@13,2 DCSAY 'File Temp' GET oMsg:file_temp PARENT oTab2 GETSIZE 80
@14,2 DCSAY 'File Size' GET oMsg:file_size PICT '99999999' PARENT oTab2
@15,2 DCSAY 'User List' GET oMsg:user_list PARENT oTab2 GETSIZE 80
@16,2 DCCHECKBOX oMsg:msg_read PROMPT 'This message has been read' PARENT oTab2
@17.5, 2 DCMULTILINE oMsg:message SIZE 95,5 NOHSCROLL FONT '10.Lucida Console' PARENT oTab2

DCGETOPTIONS SAYRIGHTBOTTOM SAYWIDTH 90 EDITPROTECT {||!lEditMode} TABSTOP
DCREAD GUI FIT OPTIONS GetOptions TITLE 'Message Maintenance' SETAPPWINDOW

CLOSE ALL
Set(_SET_DELETED,lDeleted)

SetAppWindow(oAppWindow)

slIsRunning := .f.

RETURN nil

* ------------

STATIC FUNCTION GroupMaint()

STATIC slIsRunning := .f.

LOCAL GetList[0], GetOptions, oTab1, oTab2, oBrowse, oToolBar, ;
      lDeleted := Set(_SET_DELETED,.t.), oGet1, oAppWindow := SetAppWindow(), ;
      oGroups, lAddMode := .f., lEditMode := .f., nSaveRec := 0

IF slIsRunning
  RETURN nil
ENDIF

slIsRunning := .t.

OpenGroupFile()

oGroups := IMGROUPS->(DC_DbRecord():new())
IMGROUPS->(DC_DbScatter(oGroups))

@ 0,0 DCBROWSE oBrowse ALIAS 'IMGROUPS' SIZE 60,15 ;
      PRESENTATION DC_BrowPres() ;
      FONT '10.Lucida Console' ;
      WHEN {||!lEditMode} ;
      ITEMMARKED {||IMGROUPS->(DC_DbScatter(oGroups))}

DCBROWSECOL FIELD IMGROUPS->group_name HEADER 'Name' WIDTH 20 PARENT oBrowse
DCBROWSECOL FIELD IMGROUPS->group_desc HEADER 'Description' WIDTH 40 PARENT oBrowse

@ 21,0 DCTOOLBAR oToolBar SIZE 70 BUTTONSIZE 12,1

DCADDBUTTON CAPTION 'New Group' PARENT oToolBar ;
   ACTION {||nSaveRec := IMGROUPS->(RecNo()), ;
             IMGROUPS->(dbGoTo(0)), ;
             IMGROUPS->(DC_DbScatter(oGroups)), ;
             lAddMode := .t., ;
             lEditMode := .t., ;
             DC_GetRefresh(GetList), ;
             SetAppFocus(oGet1)} ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Edit Group' PARENT oToolBar ;
   ACTION {||nSaveRec := IMGROUPS->(RecNo()), ;
             lEditMode := .t., ;
             DC_GetRefresh(GetList), ;
             SetAppFocus(oGet1) } ;
   WHEN {||!lEditMode}

DCADDBUTTON CAPTION 'Save' PARENT oToolBar ;
   ACTION {||IMGROUPS->(DC_DbGather(oGroups,lAddMode)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList)} ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Cancel' PARENT oToolBar ;
   ACTION {||IMGROUPS->(dbGoTo(nSaveRec)), ;
             IMGROUPS->(DC_DbScatter(oGroups)), ;
             lAddMode := .f., ;
             lEditMode := .f., ;
             DC_GetRefresh(GetList) } ;
   WHEN {||lEditMode}

DCADDBUTTON CAPTION 'Delete Group' PARENT oToolBar ;
   ACTION {||IIF(DC_MsgBox(,,{'Delete this Group?'},,,,.t.), ;
             (IMGROUPS->(DC_RecLock()), ;
              IMGROUPS->(DC_Blank()), ;
              IMGROUPS->(DC_RecLock()), ;
              IMGROUPS->(dbDelete()), ;
              IMGROUPS->(dbRUnlock()), ;
              IMGROUPS->(dbSkip(-1))),nil), ;
             DC_GetRefresh(GetList)} ;
   WHEN {||!lEditMode}

@ 17,0 DCSAY 'Group Name' GET oGroups:group_name PICTURE '@!' GETOBJECT oGet1
@ 18,0 DCSAY 'Group Desc' GET oGroups:group_desc

DCGETOPTIONS SAYRIGHTBOTTOM SAYWIDTH 75 EDITPROTECT {||!lEditMode} TABSTOP
DCREAD GUI FIT OPTIONS GetOptions TITLE 'Group Maintenance' SETAPPWINDOW

CLOSE ALL
Set(_SET_DELETED,lDeleted)

SetAppWindow(oAppWindow)

slIsRunning := .f.

RETURN nil

* ------------

STATIC FUNCTION PopGroups( cGroups )

LOCAL i, lStatus, aGroups, GetList[0], GetOptions, oBrowse

OpenGroupFile()
IMGROUPS->(dbGoTop())

cGroups := Strtran(cGroups,',   ','')
IF Right(cGroups,1) == ','
  cGroups := Substr(cGroups,1,Len(cGroups)-1)
ENDIF

aGroups := DC_TokenArray(cGroups,',')

@ 0,0 DCPUSHBUTTON CAPTION 'ALL Groups' SIZE 25,1  ;
      ACTION {||aGroups := {'ALL'}, ;
                DC_ReadGuiEvent(DCGUI_EXIT_OK,GetList)}

@ 1.5,0 DCSAY 'Use ENTER or DOUBLE-CLICK to tag/untag Groups' SAYSIZE 0

@ 3,0 DCBROWSE oBrowse ALIAS 'IMGROUPS' SIZE 55,20 FIT ;
      PRESENTATION DC_BrowPres() ;
      FONT '8.Lucida Console' ;
      COLOR {||IIF(Trim(IMGROUPS->group_name) $ aGroups,{nil,GRA_CLR_YELLOW},{nil,GRA_CLR_WHITE})} ;
      ITEMSELECTED ;
         {|n|n:=Ascan(aGroups,Trim(IMGROUPS->group_name)), ;
             IIF(n>0,ARemove(aGroups,n),AAdd(aGroups,Trim(IMGROUPS->group_name))), ;
             oBrowse:refreshCurrent()}

DCBROWSECOL FIELD IMGROUPS->group_name HEADER 'Name' WIDTH 15 PARENT oBrowse
DCBROWSECOL FIELD IMGROUPS->group_desc HEADER 'Description' WIDTH 40 PARENT oBrowse

DCGETOPTIONS NORESIZE TABSTOP

DCREAD GUI ;
   FIT ;
   ADDBUTTONS ;
   NOAUTORESTORE ;
   TO lStatus ;
   TITLE 'Tag Group List' ;
   MODAL ;
   SETAPPWINDOW ;
   OPTIONS GetOptions

IF lStatus
  cGroups := ''
  FOR i := 1 TO Len(aGroups)
    cGroups += Alltrim(aGroups[i]) + IIF(i<Len(aGroups),',','')
  NEXT
ENDIF

RETURN cGroups

* ------------

STATIC FUNCTION ViewClients()

LOCAL aClients[0], i, oClient

FOR i := 1 TO Len(saClientList)
  oClient := saClientList[i]
  IF .f. // Valtype(oClient) == 'O' .AND. oClient:isDerivedFrom('DC_WebSocketHandler') .AND. ;
    AAdd(aClients,oClient:requestClient:varList)
  ELSE
    AAdd(aClients,saClientList[i])
  ENDIF
NEXT

DC_ArrayView(aClients)

RETURN nil

* ------------

STATIC FUNCTION _QuitApp(oBrowse)

LOCAL lStatus := .T.

IF !Empty(saClientList)
  IF DC_MsgBox(,,{'There are ' + Alltrim(Str(Len(saClientList))) + ' clients connected.', ;
               'Are you sure you want to exit?'},,,,.t.)
    lStatus := .T.
  ELSE
    lStatus := .F.
  ENDIF
ENDIF

IF lStatus
  IIF(Valtype(soServer)=='O',LogActivity(oBrowse,'Server  Stopped'),nil)
ENDIF

RETURN lStatus

* -------------

STATIC PROCEDURE LogError( oOwner, nErrCode )

Local i, cError, cProcName, oThread

   cError := "= IM Server Error =======" + CRLF +;
             "Date: " + DTOS(Date()) + ", Time: " + Time() + CRLF +;
             "App: " + AppName(.t.) + ", Thread: " + LTrim(Str(ThreadID())) + CRLF

   if ValType(oOwner) == "O"
      if oOwner:isDerivedFrom('Error')
         // error generated while processing http server request
         cError += if(oOwner:severity > ES_WARNING, "ERROR ", "WARNING " ) +;
                   if(ValType(oOwner:subSystem)   == "C", oOwner:subSystem, "???" ) + "/" +;
                   if(ValType(oOwner:subCode)     == "N", NTrim(oOwner:subCode), "???" ) + CRLF +;
                   if(ValType(oOwner:cargo)       == "C", oOwner:cargo + CRLF, "" ) +;
                   if(ValType(oOwner:description) == "C", oOwner:description, "" )

         // add either filename or operation
         if !Empty(oOwner:filename)
            cError += ": " + oOwner:filename
         elseif !Empty(oOwner:operation)
            cError += ": " + oOwner:operation
         endif

         if !Empty( oOwner:osCode )
            cError += CRLF + "(OS Error: " + NTrim(oOwner:osCode) + ")"
         endif
      else
         // some kind of socket related error
         cError += "WinSock error: " + NTrim(nErrCode) + " - " + oOwner:ErrorText(nErrCode)
      endif
   endif

   for i := 2 to 12
      if Empty(cProcName := ProcName(i))
         exit
      endif
      cError += CRLF+"Called from "+cProcName+"("+LTrim(Str(ProcLine(i)))+")"
   next

   SaveToFile( cError + CRLF, ERR_LOG_FILE )

oThread := ThreadObject()
oThread:httpResponse:ContentType('text/plain')

cError := '<pre>' + cError + '</pre>'

Break( cError )

Return

* ------------

STATIC FUNCTION SaveToFile( cText, cFileName )
   Local i := 0, nHandle

   if ValType(cText) != "C" .or. Empty(cText)
      Return .f.
   endif

   if FExists(cFileName)
      while (nHandle := FOpen(cFileName, FO_WRITE + FO_DENYWRITE)) < 0
         if ++i > 9
            Return .f.
         endif
         Sleep(20)
      end
      FSeek(nHandle, 0, FS_END)  // position to end of file
   else
      while (nHandle := FCreate(cFileName)) < 0
         if ++i > 9
            Return .f.
         endif
         Sleep(20)
      end
   endif

   FWrite(nHandle, cText + CRLF)
   FClose(nHandle)

   Return .t.

// A websocket handler must implement the methods
// :onConnect(), onDisconnect(), onText() and
// :onBinary().
//
// The methods are triggered from the HttpEndpoint.
//
CLASS IMServer FROM DC_WebSocketHandler

EXPORTED:
INLINE METHOD OnConnect

// wtf 'Connect' COLOR GRA_CLR_DARKGREEN WINDOW 1

AAdd(saClientList,self)
::DC_WebSocketHandler:requestHandler := {|oRequest|RequestHandler(oRequest,self)}
RETURN ::DC_WebSocketHandler:onConnect()

INLINE METHOD OnDisconnect
LOCAL nFound
nFound := AScan(saClientList,self)
IF nFound > 0
  ARemove(saClientList,nFound)
ENDIF

// wtf 'DisConnect' COLOR GRA_CLR_DARKRED WINDOW 1

RETURN ::DC_WebSocketHandler:onDisconnect()

INLINE METHOD OnBinary( cBinary )
LOCAL oRequest := Bin2Var( cBinary )
// WTF oRequest:action + ' ' + DC_XtoC(oRequest:getVar('UserId')) COLOR GRA_CLR_DARKCYAN WINDOW 1
RETURN ::DC_WebSocketHandler:onBinary(cBinary)

ENDCLASS

* ----------

FUNCTION RequestHandler( oRequest, oHandler )

   Local cAction, oResponse, cString, cEnvelope, cToUser, cFromUser, ;
         cPassword, cUserId, cMessage, cMessage2, nStatus := 0, cFromName, ;
         cTempFileName := '', cTo, cFrom, oSession, oThread := ThreadObject()

   // if the request is parsed correctly, this should be the action requested by the client
   cAction   := oRequest:Action


// wtf oRequest COLOR GRA_CLR_BROWN WINDOW 1

   oHandler:requestClient := oRequest

   // create another object where we will write our response
   oResponse := DC_WebSocketData():new()

   // tack on the word "Response" to the reply envelope action, eg. "GetTimeResponse"
   oResponse:Action := cAction + "Response"

   // convert to uppercase so we are not case sensitive
   cAction   := Upper(cAction)

   cMessage := 'Unknown'

   BEGIN SEQUENCE

   cUserID := Alltrim(DC_XtoC(oRequest:getVar('UserID')))
   cPassword := Alltrim(DC_XtoC(oRequest:getVar('Password')))
   cFromUser := Alltrim(DC_XtoC(oRequest:getVar('FromUser')))

   cTo := cUserID
   IF Empty(cTo)
     cTo := DC_Array2String(oRequest:getVar('ToUserList'))
   ENDIF
   cFrom := cFromUser
   IF Empty(cFrom)
     cFrom := cUserID
   ENDIF

   IF cAction == 'GETFILE'
     cFrom := DC_XtoC(oRequest:getVar('Filename'))
   ENDIF

   LogActivity( soBrowse, cAction + ': ' + cFrom )

   IF IsFunction(cAction)
     nStatus := &(cAction)(oRequest,oResponse,cUserId,cPassword,cFromUser,oThread,oSession)
     BREAK
   ELSE
     oResponse:SetFault("Client", "Client request " + cAction + " not available")
   ENDIF

   // attach the SOAP response envelope to the HTTP response object
   // the Xb2.NET HTTP server will take care of the rest

   RECOVER USING cMessage2

   IF cMessage2 # nil
     cMessage := cMessage2
     nStatus := 900
     wtf cMessage // don't remove this.  displays in debug window if a runtime error occurs
   ENDIF

   END SEQUENCE

   IF nStatus < 900
     cMessage := LookupMessage(nStatus)
   ENDIF
   oResponse:setVar('StatusMessage', cMessage )
   oResponse:setVar('Status', nStatus )

   CLOSE ALL

   DC_GetRefresh(saGetList,'USERS')

Return oResponse

* ========== Misc Functions =============

FUNCTION GetDateTime( oRequest, oResponse )

oResponse:SetVar("Date", Date())
oResponse:SetVar("Time", Time())

RETURN 213

* ======== Log On/Off Functions =========

FUNCTION LogOn( oRequest, oResponse, cUserID, cPassword )

LOCAL nStatus, nFound, cUserAlias, lVisible, cLanguage, ;
      aClient, aMedia, aPhotos, cClientVersion

aClient := Directory(DC_Path(AppName(.t.)) + 'imclient\imclient.dll')
aMedia := Directory(DC_Path(AppName(.t.)) + 'media\*.*')
aPhotos := Directory(DC_Path(AppName(.t.)) + 'photos\*.*')

nStatus := _ValidateUser( @cUserId, cPassword, .T. )

IF nStatus > 0
  RETURN nStatus
ENDIF

cUserAlias := oRequest:getVar('Alias')
lVisible := oRequest:getVar('Visible')
cLanguage := oRequest:getVar('Language')
cClientVersion := oRequest:getVar('ClientVersion')
oResponse:setVar('Client',aClient)
oResponse:setVar('Media',aMedia)
oResponse:setVar('Photos',aPhotos)

DEFAULT lVisible := .t.

IF IMUSERS->(DC_RecLock(5))
  IF !Empty(cUserAlias) .AND. ;
    !(Upper(Alltrim(cUserAlias)) == Upper(Alltrim(IMUSERS->user_alias)))
    IMUSERS->user_alias := cUserAlias
  ENDIF
  IF !Empty(cLanguage)
    IMUSERS->language := cLanguage
  ENDIF
  IF !Empty(cClientVersion)
    IMUSERS->client_ver := cClientVersion
  ENDIF
  IMUSERS->(dbUnlock())
ENDIF

scUserId := cUserId
UserFlag( cUserID, nil, lVisible )
UserFlag( '$$$', 301, nil, cUserId )

RETURN GetUsers( oRequest, oResponse, cUserId, cPassword )

* -----------

FUNCTION LogOff( oRequest, oResponse, cUserID, cPassword )

LOCAL nStatus, nFound

nStatus := _ValidateUser( @cUserId, cPassword, .T. )

IF nStatus > 0
  RETURN nStatus
ENDIF

cUserId := Alltrim(Upper(cUserId))
nFound := AScan( saUsers, {|a|Upper(a[1])==Upper(cUserID)})

UserFlag( '$$$', 302 )

IF nFound > 0
  ARemove( saUsers, nFound )
ENDIF

RETURN 302

* ========== User Functions =============

FUNCTION GetUsers( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, aUsers[0], lUserActive, lActiveOnly, lUserVisible

nStatus := _ValidateUser( @cUserId, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF
lActiveOnly := oRequest:getVar('OnlineOnly')
DEFAULT lActiveOnly := .f.

IMUSERS->(dbGoTop())
DO WHILE !IMUSERS->(Eof())
  lUserActive := IsUserActive( IMUSERS->user_id )
  lUserVisible := IsUserVisible( IMUSERS->user_id )
  IF IMUSERS->mode == 1 .OR. !lUserVisible // Invisible
    IMUSERS->(dbSkip())
    LOOP
  ELSEIF lActiveOnly .AND. !lUserActive
    IMUSERS->(dbSkip())
    LOOP
  ELSEIF Upper(Alltrim(IMUSERS->user_id)) == Alltrim(Upper(cUserId))
    IMUSERS->(dbSkip())
    LOOP
  ELSEIF Empty(IMUSERS->user_id) .OR. IMUSERS->(Deleted())
    IMUSERS->(dbSkip())
    LOOP
  ENDIF
  AAdd( aUsers, { ;
        Trim(IMUSERS->user_id), ;
        Trim(IMUSERS->user_last), ;
        Trim(IMUSERS->user_first), ;
        Trim(IMUSERS->user_alias), ;
        Trim(IMUSERS->user_email), ;
        Trim(IMUSERS->status), ;
        IMUSERS->mode, ;
        .f., ;
        Trim(IMUSERS->user_group), ;
        Trim(IMUSERS->client_ver) } )
  Atail(aUsers)[8] := lUserActive
  IF !lUserActive
    Atail(aUsers)[6] := ''
  ELSE
    Atail(aUsers)[6] := 'Available'
  ENDIF
  IMUSERS->(dbSkip())
ENDDO

oResponse:setVar('UserList',aUsers)
oResponse:setVar('HostAlias',scHostAlias)
GetGroups( oRequest, oResponse, cUserId, cPassword, .f. )

RETURN 200

* --------------

FUNCTION DeleteUser( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cAuthID, cAuthPassword

cAuthID := oRequest:getVar('AuthID')
cAuthPassword := oRequest:getVar('AuthPassword')

nStatus := _ValidateUser( @cAuthID, cAuthPassword, .f. )
IF nStatus > 0
  RETURN 111
ENDIF
OpenUserFile()
IMUSERS->(dbSeek(Upper(cUserID)))
IF !IMUSERS->(Found())
  RETURN 105
ENDIF
IF !IMUSERS->(DC_Blank())
  RETURN 105
ENDIF
UserFlag( '$$$', 201 )

RETURN 201

* --------------

FUNCTION UpdateUser( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cAuthId, cAuthPassword, cLastName, cFirstName, ;
      cAlias, cEmail

cAuthID := oRequest:getVar('AuthID')
cAuthPassword := oRequest:getVar('AuthPassword')

nStatus := _ValidateUser( @cAuthID, cAuthPassword, .f. )
IF nStatus > 0
  RETURN 111
ENDIF
cLastName := oRequest:getVar('LastName')
cFirstName := oRequest:getVar('FirstName')
cAlias := oRequest:getVar('Alias')
cEmail := oRequest:getVar('Email')
OpenUserFile()
IMUSERS->(dbSeek(Upper(cUserID)))
IF !IMUSERS->(Found())
  RETURN 105
ENDIF
IF !IMUSERS->(DC_RecLock(5))
  RETURN 150
ENDIF
IMUSERS->user_id := cUserID
IMUSERS->password := cPassword
IMUSERS->user_last := cLastName
IMUSERS->user_first := cFirstName
IMUSERS->user_alias := cAlias
IMUSERS->user_email := cEmail
IMUSERS->(dbRUnlock())
UserFlag( '$$$', 205 )

RETURN 205

* --------------

FUNCTION GetUserPreferences( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, aPreferences

nStatus := _ValidateUser( @cUserId, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF
OpenUserFile()
IMUSERS->(dbSeek(Upper(cUserID)))
IF !IMUSERS->(Found())
  RETURN 105
ENDIF
aPreferences := DC_Str2Ar(IMUSERS->preference)
DEFAULT aPreferences := {}
ASize(aPreferences,1)
DEFAULT aPreferences[1] := 0

oResponse:setVar('DisplayMode',aPreferences[1])
oResponse:setVar('Language',IMUSERS->language)

RETURN 200

* --------------

FUNCTION SetUserPreferences( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, aPreferences, nDisplayMode, cLanguage

nStatus := _ValidateUser( @cUserId, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF
OpenUserFile()
IMUSERS->(dbSeek(Upper(cUserID)))
IF !IMUSERS->(Found())
  RETURN 105
ENDIF
IF !IMUSERS->(DC_RecLock(5))
  RETURN 150
ENDIF
nDisplayMode := oRequest:getVar('DisplayMode')
DEFAULT nDisplayMode := 1

cLanguage := oRequest:getVar('Language')
aPreferences := { nDisplayMode }

IMUSERS->preference := DC_Ar2Str(aPreferences)
IF !Empty(cLanguage)
  IMUSERS->language := cLanguage
ENDIF

IMUSERS->(dbRUnlock())

RETURN 200

* --------------

FUNCTION AddUser( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cAuthId, cAuthPassword, cLastName, cFirstName, ;
      cAlias, cEmail, lUpdate := .f.

cAuthID := oRequest:getVar('AuthID')
cAuthPassword := oRequest:getVar('AuthPassword')

nStatus := _ValidateUser( @cAuthID, cAuthPassword, .f. )
IF nStatus > 0
  RETURN 111
ENDIF
cLastName := oRequest:getVar('LastName')
cFirstName := oRequest:getVar('FirstName')
cAlias := oRequest:getVar('Alias')
cEmail := oRequest:getVar('Email')
OpenUserFile()
IMUSERS->(dbSeek(Upper(cUserID)))
IF IMUSERS->(Found())
  IF Alltrim(IMUSERS->password) == Alltrim(cPassword)
    lUpdate := .t.
  ELSE
    RETURN 106
  ENDIF
ENDIF
IF !lUpdate
  IMUSERS->(dbSeek('   '))
  IF IMUSERS->(Found()) .AND. !IMUSERS->(DC_RecLock(5))
    RETURN 150
  ELSEIF !IMUSERS->(DC_AddRec(5))
    RETURN 150
  ENDIF
ENDIF
IF lUpdate
  IMUSERS->(DC_RecLock(5))
ENDIF
IMUSERS->user_id := cUserID
IMUSERS->password := cPassword
IMUSERS->user_last := cLastName
IMUSERS->user_first := cFirstName
IMUSERS->user_alias := cAlias
IMUSERS->user_email := cEmail
IMUSERS->(dbRUnlock())
IF lUpdate
  nStatus := 211
ELSE
  nStatus := 210
ENDIF
UserFlag( '$$$', nStatus )

RETURN nStatus

* -------------

FUNCTION ValidateUser( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cUserAlias, aPhoto, aDir

nStatus := _ValidateUser( @cUserId, cPassword, .f. )
IF nStatus > 0
  RETURN nStatus
ENDIF

IF nStatus == -1 // ADMIN login
  oResponse:setVar('LastName','Administrator')
  oResponse:setVar('FirstName','System')
  oResponse:setVar('Alias','Admin')
  oResponse:setVar('Email','rogerdonnay@donnay-software.com')
ELSE
  oResponse:setVar('LastName',Alltrim(IMUSERS->user_last))
  oResponse:setVar('FirstName',Alltrim(IMUSERS->user_first))
  oResponse:setVar('Alias',Alltrim(IMUSERS->user_alias))
  oResponse:setVar('Email',Alltrim(IMUSERS->user_email))
  aDir := Directory(DC_Path(AppName(.t.)) + 'Photos\' + cUserId + '.JPG')
  IF !Empty(aDir)
    aPhoto := { cUserId+'.JPG', aDir[1,2]}
    oResponse:setVar('Photo',aPhoto)
  ENDIF
ENDIF
oResponse:setVar('Active',IsUserActive(cUserID))
oResponse:setVar('Photo',aPhoto)

// UserFlag( '$$$', 206 )

RETURN 206

* -------------

FUNCTION SetUserStatus( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cStatus, nMode

nStatus := _ValidateUser( @cUserID, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF

cStatus := DC_Xtoc(oRequest:GetVar("Status"))
nMode := Val(DC_XtoC(oRequest:GetVar('Mode')))
IF !OpenUserFile()
  RETURN 150
ENDIF
IF !IMUSERS->(DC_RecLock(5))
  RETURN 150
ENDIF
IMUSERS->status := cStatus
IMUSERS->mode := nMode
IMUSERS->(dbRUnlock())
// UserFlag( '$$$', 211 )

RETURN 211

* ========== Group Functions =============

FUNCTION GetGroups( oRequest, oResponse, cUserId, cPassword, lValidate )

LOCAL nStatus, aGroups[0]

DEFAULT lValidate := .t.

IF lValidate
  nStatus := _ValidateUser( @cUserId, cPassword, .T. )
  IF nStatus > 0
    RETURN nStatus
  ENDIF
ENDIF

IF !OpenGroupFile()
  RETURN 105
ENDIF
aGroups := {}
DO WHILE !IMGROUPS->(Eof())
  IF !Empty(IMGROUPS->group_name) .AND. !IMGROUPS->(Deleted())
    AAdd( aGroups, { ;
          Trim(IMGROUPS->group_name), ;
          Trim(IMGROUPS->group_desc), ;
          Trim(IMGROUPS->group_wave) } )
  ENDIF
  IMGROUPS->(dbSkip())
ENDDO

oResponse:setVar('GroupList',aGroups)
oResponse:setVar('HostAlias',scHostAlias)

RETURN 212

* ========== Message Functions =============

FUNCTION GetMessage( oRequest, oResponse, cUserId, cPassword )

LOCAL nStatus, cToLanguage, cFromLanguage

nStatus := _ValidateUser( @cUserId, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF
cToLanguage := IMUSERS->language
OpenMessageFile()
IMMSG->(OrdSetFocus('UNREAD_MSG'))
IMMSG->(dbSeek(Upper(cUserId)))
IF IMMSG->(Eof())
  RETURN 203 // no messages
ENDIF
OpenUserFile()
IMUSERS->(dbSeek(IMMSG->from_user))
cFromLanguage := IMUSERS->language
oResponse:SetVar('FromLastName',Trim(IMUSERS->user_last))
oResponse:SetVar('FromFirstName',Trim(IMUSERS->user_first))
oResponse:SetVar('FromAlias',Trim(IMUSERS->user_alias))
oResponse:SetVar('FromEmail',Trim(IMUSERS->user_email))
oResponse:SetVar('Message',ConvertLanguage(Trim(IMMSG->message),cToLanguage,cFromLanguage))
oResponse:SetVar('FromUser',Trim(IMMSG->from_user))
oResponse:SetVar('DateSent',IMMSG->date)
oResponse:SetVar('TimeSent',IMMSG->time)
oResponse:SetVar('Font',Alltrim(IMMSG->font))
oResponse:SetVar('FGColor',IMMSG->fgcolor)
oResponse:SetVar('BGColor',IMMSG->bgcolor)
oResponse:SetVar('FileName',Alltrim(IMMSG->file_name))
oResponse:SetVar('TempFileName',Alltrim(IMMSG->file_temp))
oResponse:SetVar('WaveFile',Alltrim(IMMSG->wave))
oResponse:SetVar('ConversationNum',IMMSG->conv_nmbr)
oResponse:SetVar('MessageType',IMMSG->msg_type)

IF snDays == 0 .OR. Trim(IMMSG->message) $ {'<<typing>>','<<nudge>>'} ;
               .OR. Empty(IMMSG->message) .OR. IMMSG->clone
  IMMSG->(DC_Blank())
ELSE
  IF IMMSG->(dbRLock())
    IMMSG->msg_read := .t.
  ENDIF
ENDIF

oResponse:SetVar('MoreMessages',IMMSG->(dbSeek(Upper(cUserId))))

RETURN 204

* -------------

STATIC FUNCTION ConvertLanguage( cMessage, cToLanguage, cFromLanguage )

RETURN cMessage

/*
LOCAL cLanguage

IF Empty(cToLanguage)
  cToLanguage := 'EN'
ENDIF
IF Empty(cFromLanguage)
  cFromLanguage := 'EN'
ENDIF

IF cToLanguage == cFromLanguage .OR. Empty(cMessage) .OR. Left(cMessage,2) == '<<'
  RETURN cMessage
ENDIF

cLanguage := cFromLanguage + '_' + cToLanguage
cMessage := BabelFish( cLanguage, cMessage )

RETURN cMessage

* ---------------

STATIC FUNCTION BabelFish( cLanguage, cText )

LOCAL oAgent := xbSOAPEnvelope():new(), oResult, aResult, cString

oAgent:namespace := "urn:xmethodsBabelFish"
oAgent:SetVar("translationmode", Lower(cLanguage))
oAgent:SetVar("sourcedata", cText)

oAgent:RecvTimeOut(10000)
oResult := oAgent:Execute("http://services.xmethods.net:80/perl/soaplite.cgi", "BabelFish")

cString := oResult:getVar('return')
IF Valtype(cString) # 'C' .OR. Empty(cString)
  cString := cText
ENDIF

RETURN cString
*/

* -------------

FUNCTION SendMessage( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL nStatus, aToUserList, nConversationNum, lActiveOnly, nMessageType

aToUserList := oRequest:getVar('ToUserList')
nConversationNum := oRequest:getVar('ConversationNum')
lActiveOnly := oRequest:getVar('ActiveOnly')
nMessageType := oRequest:getVar('MessageType')

DEFAULT lActiveOnly := .f.

nStatus := _ValidateUser( @cFromUser, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF

RETURN _SendMessage( aToUserList, cFromUser, oRequest, oResponse,, ;
                     nConversationNum, lActiveOnly, nMessageType )

* -------------

FUNCTION SendGroupMessage( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL nStatus, cGroupName, aToUserList, cGroupList, nConversationNum, ;
      lActiveOnly, nMessageType

nStatus := _ValidateUser( @cFromUser, cPassword, .T. )
IF nStatus > 0
  RETURN nStatus
ENDIF

OpenGroupFile()
OpenUserFile()
cGroupName := Upper(Alltrim(DC_XtoC(oRequest:getVar('GroupName'))))
nConversationNum := oRequest:getVar('ConversationNum')
lActiveOnly := oRequest:getVar('OnlineOnly')
nMessageType := oRequest:getVar('MessageType')
IMGROUPS->(dbSeek(cGroupName))
IMUSERS->(DC_DbGoTop())
aToUserList := {}
DO WHILE !IMUSERS->(Eof()) .AND. !Empty(cGroupName)
  cGroupList := ',' + Upper(Alltrim(IMUSERS->user_group)) + ','
  cGroupList := Strtran(cGroupList,' ','')
  IF (',' + cGroupName + ',' $ cGroupList .OR. cGroupName = 'ALL-STATIONS') .AND. ;
    !(Alltrim(Upper(cFromUser)) == Alltrim(Upper(IMUSERS->user_id)))
    AAdd( aToUserList,Alltrim(Upper(IMUSERS->user_id)))
  ENDIF
  IMUSERS->(DC_DbSkip())
  Sleep(1)
ENDDO

RETURN _SendMessage( aToUserList, cFromUser, oRequest, oResponse, ;
                     Trim(IMGROUPS->group_wave), nConversationNum, ;
                     lActiveOnly, nMessageType )

* -------------

FUNCTION WaitForMessage( oRequest, oResponse, cUserId, cPassword, cFromuser, oThread, oSession )

LOCAL nStatus, dDate, nTimeOut, nSeconds, nUserStatus, nSleep, nFound

nStatus := _ValidateUser( @cUserId, cPassword, .t. )

cUserId := Alltrim(Upper(cUserId))

IF nStatus > 0
  RETURN nStatus
ENDIF

CLOSE DATABASES

dDate := Date()
nStatus := 112
nSeconds := Seconds()

ThreadObject():cargo := cUserID

nFound := AScan(saUsers,{|a|a[1]==cUserId})
IF nFound > 0
  saUsers[nFound,3] := Seconds()
ENDIF

DO WHILE Seconds() - nSeconds < snTimeOut .AND. dDate == Date()
  IF (nFound := AScan(saUsers,{|a|Upper(a[1])==Upper(cUserId)})) == 0
    nStatus := 303
    EXIT
  ENDIF
  nUserStatus := UserFlag( cUserId ) // .OR. Empty(saUsers)
  IF nUserStatus > 0
    UserFlag(cUserId,0)
    nStatus := nUserStatus // You have messages or user status has changed
    EXIT
  ENDIF
  Sleep(100)
ENDDO

IF nStatus == 207 // You have a message
  nStatus := GetMessage( oRequest, oResponse, cUserId, cPassword )
ELSEIF nStatus == 303 // Forced log off
ELSEIF nStatus == 304 // Forced shut down
ELSE   // User status has changed
  nSleep := AScan(saUsers,{|a|Upper(a[1])==Upper(cUserId)})
  IF Valtype(nSleep) = 'N'
    Sleep(nSleep*100)
  ENDIF
  IF nStatus == 301 .AND. !Empty(scUserId) // user logged on
    oResponse:setVar('NewUser',scUserId)
  ENDIF
  nStatus := GetUsers( oRequest, oResponse, cUserId, cPassword )
ENDIF

RETURN nStatus

* ========== Conversation Functions =============

FUNCTION GetConversationList( oRequest, oResponse, cUserId, cPassword )

LOCAL aConversation[0], nStatus

OpenMessageFile()
OpenUserFile()

IMMSG->(OrdSetFocus('CONV_UNIQ'))

DO WHILE !IMMSG->(Eof())
  AAdd( aConversation, ;
        { IMMSG->CONV_NMBR, ;
          IMMSG->DATE, ;
          IMMSG->TIME, ;
          IMMSG->USER_LIST, ;
          UserNames( IMMSG->USER_LIST) } )
  IMMSG->(dbSkip())
ENDDO

oResponse:setVar( 'Conversations', aConversation )

RETURN nStatus

* -----------

STATIC FUNCTION UserNames( cUserList )

LOCAL i, aTokens := DC_TokenArray( cUserList, ',' ), cUserNames := ''

FOR i := 1 TO Len(aTokens)
  IMUSERS->(dbSeek(Alltrim(Upper(aTokens[i]))))
  cUserNames += Alltrim(IMUSERS->user_first) + ' ' + Alltrim(IMUSERS->user_last)
  IF i < Len(aTokens)
    cUserNames += ','
  ENDIF
NEXT

RETURN cUserNames

* ========== File Transfer Functions =============

FUNCTION GetFile( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL cUserName, cFileName, cPath, nSegmentSize, nSegment, aDir, ;
      cBuffer, dFileDate, cFileTime, nBytesInChunk, nHandle, ;
      nFileSize, cTerminator, nStartByte, i, cString, lCompress, ;
      cOutBuffer, nOutLen, nStatus, cLocation, lDeleteFile

cUserName := oRequest:GetVar('UserName')
cFileName := oRequest:GetVar('FileName')
cLocation := oRequest:GetVar('Location')

IF !Empty(cLocation)
  cPath := DC_Path(Appname(.t.)) + cLocation + '\'
  cFileName := cPath + cFileName
ELSEIF !Empty(cUserName) .AND. !('\' $ cFileName)
  cPath := DC_Path(AppName(.t.)) + 'USERS\' + cUserName + '\'
  cFileName := cPath + cFileName
ENDIF

nSegmentSize := oRequest:GetVar('SegmentSize')
nSegment := oRequest:GetVar('Segment')
nStartByte := oRequest:GetVar('StartByte')
lCompress := oRequest:GetVar('Compress')
aDir := Directory(cFileName)
nHandle := FOpen(cFileName)
nFileSize := FSize(nHandle)

cTerminator := oRequest:GetVar('Terminator')
lDeleteFile := oRequest:GetVar('DeleteFile')

IF Valtype(nStartByte) == 'N'
  FSeek( nHandle, nStartByte )
ELSEIF Valtype(nSegment) == 'N'
  FSeek( nHandle, nSegmentSize*(nSegment-1) )
ENDIF
cBuffer := Space(nSegmentSize)
nBytesInChunk := Fread( nHandle, @cBuffer, nSegmentSize )

cBuffer := Substr(cBuffer,1,nBytesInChunk)

FClose( nHandle )
IF Valtype(cTerminator) == 'C'
  FOR i := Len(cBuffer) TO 1 STEP -1
    cString := Substr(cBuffer,i,Len(cTerminator))
    IF cString == cTerminator
      cBuffer := Substr(cBuffer,1,i-1) + cTerminator
      nBytesInChunk := Len(cBuffer)
      EXIT
    ENDIF
  NEXT
ENDIF

IF Len(aDir) == 1
  dFileDate := aDir[1,F_DATE]
  cFileTime := aDir[1,F_TIME]
ELSE
  dFileDate := Date()
  cFileTime := Time()
ENDIF
DEFAULT lCompress := .f.
IF lCompress
  cOutBuffer := Space((nBytesInChunk*1.02)+12)
  nOutLen := Len(cOutBuffer)
  nStatus := Compress( @cOutBuffer, @nOutLen, @cBuffer, nBytesInChunk )
  IF nStatus == 0
    cBuffer := SubStr(cOutBuffer,1,nOutLen)
    nBytesInChunk := nOutLen
    oResponse:setVar( 'Compressed', .t. )
  ELSE
    oResponse:setVar( 'Compressed', .f. )
  ENDIF
ELSE
  oResponse:setVar( 'Compressed', .f. )
ENDIF

oResponse:setVar( "BufferChunk", cBuffer )
oResponse:setVar( "Message", 'Thank you for using IM File Transfer' )
oResponse:setVar( "BytesInChunk", nBytesInChunk )
oResponse:setVar( "FileSize", nFileSize )
oResponse:setVar( "FileDate", dFileDate )
oResponse:setVar( "FileTime", cFileTime )

IF Valtype(lDeleteFile) == 'L' .AND. lDeleteFile .AND. ;
    nBytesInChunk > 0 .AND. nBytesInChunk < nSegmentSize
  Ferase(cFileName)
ENDIF

RETURN 0

* -------------

FUNCTION GetHostTempFileName( oRequest, oResponse )

LOCAL cLocation, cExtension, cFileName

cLocation := oRequest:getVar('Location')
cExtension := oRequest:getVar('Extension')
cFileName := TempFileName( DC_Path(AppName(.t.)) + cLocation, cExtension )

oResponse:setVar('FileName',DC_Path(cFileName,.t.))

RETURN 214

* -------------

FUNCTION PutFile( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL cUserName, cFileName, nFileSize, nHandle, cBuffer, nBytesInChunk, ;
      nSegmentSize, nSegment, cPath, dFileDate, cFileTime, cTempFileName, ;
      cLocation, lMakeTempFile, cExtension

cUserName := oRequest:GetVar('UserName')
cFileName := oRequest:GetVar('FileName')
nFileSize := oRequest:GetVar('FileSize')
cLocation := oRequest:GetVar('Location')

lMakeTempFile := oRequest:GetVar('MakeTempFile')
cExtension := oRequest:GetVar('Extension')
cBuffer := oRequest:GetVar('BufferChunk')
nBytesInChunk := oRequest:GetVar('BytesInChunk')
nSegmentSize := oRequest:GetVar('SegmentSize')
nSegment := oRequest:GetVar('Segment')

IF nSegment == 0 // get full path of file to be created and return to client
  IF !Empty(cLocation)
    cPath := DC_Path(Appname(.t.)) + cLocation + '\'
  ELSEIF !Empty(cUsername)
    cPath := DC_Path(AppName(.t.)) + 'USERS\' + cUserName + '\'
    CreateDirectoryA( cPath )
  ENDIF
  IF lMakeTempFile
    cFileName := cPath + TempFileName( cPath, cExtension )
  ELSE
    cFileName := cPath + cFileName
  ENDIF
  oResponse:setVar("FileName", cFileName) // store full path of file name to return to client
  RETURN 0
ELSE
  // FileName must contain a full path
  IF nSegment == 1
    nHandle := FCreate(cFileName,FC_NORMAL)
  ELSE
    nHandle := FOpen(cFileName,FO_READWRITE)
  ENDIF
ENDIF

IF nHandle > 0
  FSeek( nHandle, 0, FS_END )
  FWrite( nHandle, cBuffer, nBytesInChunk )
  FClose( nHandle )
  oResponse:setVar( "Message", 'File Segment Received' )
  oResponse:setVar( "Status", 0 )
ELSE
  oResponse:setVar( "Message", 'File Could not be created: ' + Alltrim(cFileName))
  oResponse:setVar( "Status", 1 )
ENDIF
dFileDate := oRequest:GetVar('FileDate')
cFileTime := oRequest:GetVar('FileTime')

DEFAULT dFileDate := Date(), ;
        cFileTime := Time()

IF !IsFunction('SetFDati') // From Xbase Tools
  DllLoad(DC_Path(AppName(.t.)) + '..\runtime\XbtBase1.dll')
ENDIF
IF IsFunction('SetFDati') // From Xbase Tools
  PRIVATE _cFileName := cFileName
  PRIVATE _dFileDate := dFileDate
  PRIVATE _cFileTime := cFileTime
  &('SetFDaTi( _cFileName, _dFileDate, _cFileTime )')
ENDIF

RETURN 0

* ------------

FUNCTION CopyFile( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL cUserName, cFromName, cFileName, cToPath, cFromPath, cFromFile, ;
      cToFile, nStatus

cUserName := oRequest:GetVar('UserName')
cFromName := oRequest:GetVar('FromName')
cFileName := oRequest:GetVar('FileName')
IF !Empty(cUserName) .AND. !Empty(cFromName)
  cToPath := DC_Path(AppName(.t.)) + 'USERS\' + cUserName + '\'
  CreateDirectoryA( cToPath )
  cFromPath := DC_Path(AppName(.t.)) + 'USERS\' + cFromName + '\'
  cFromFile := cFromPath + cFileName
  cToFile := cToPath + cFileName
  IF !File(cFromFile)
    nStatus := 110
  ELSE
    COPY FILE (cFromFile) TO (cToFile)
    nStatus := 208
  ENDIF
ELSE
  nStatus := 109
ENDIF

RETURN nStatus

* ------------

FUNCTION DeleteFile( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL nStatus, cUserName, cFileName, cPath, cLocation

cUserName := oRequest:GetVar('UserName')
cFileName := oRequest:GetVar('FileName')
cLocation := oRequest:GetVar('Location')

IF !Empty(cLocation)
  cPath := DC_Path(AppName(.t.)) + cLocation + '\'
  cFileName := cPath + cFileName
ELSEIF !Empty(cUserName)
  cPath := DC_Path(AppName(.t.)) + 'USERS\' + cUserName + '\'
  cFileName := cPath + cFileName
ELSE
  nStatus := 109
ENDIF

IF !Empty(nStatus)
  IF !File(cFileName)
    nStatus := 110
  ELSE
    FErase(cFileName)
    nStatus := 208
  ENDIF
ENDIF

RETURN nStatus

* -----------

FUNCTION GetDirectory( oRequest, oResponse, cUserId, cPassword, cFromuser )

LOCAL cUserName, cPath, cWildCard, aDir, i, nStart, nEnd, cText, ;
      nStatus := 0, nFound

cUserName := oRequest:GetVar('UserName')
cPath := oRequest:GetVar('Path')
cWildCard := oRequest:GetVar('WildCard')
IF !Empty(cUserName) .AND. Empty(cPath)
  cPath := DC_Path(AppName(.t.)) + 'USERS\' + cUserName + '\'
  CreateDirectoryA( cPath )
ENDIF

DEFAULT cWildCard := '*.*'

aDir := Directory(cPath + '\' + cWildCard)
FOR i := 1 TO Len(aDir)
  ASize( aDir[i], 5 )
  cText := MemoRead( cPath + '\' + aDir[i,1] )
  IF (nFound := At( '..DESCRIPTION',Upper(cText) )) > 0
    nStart := At( CRLF, cText, nFound ) + 2
    nEnd := At( '..END', Upper(cText), nFound )
    cText := Substr( cText, nStart, nEnd-nStart )
  ELSE
    cText := ''
  ENDIF
  aDir[i,5] := Strtran(cText,CRLF,'')
NEXT

oResponse:setVar( "Directory", aDir )

RETURN nStatus

* -------------

STATIC FUNCTION LookupMessage( nStatus )

LOCAL aMessage, nFound

DEFAULT nStatus := 0

aMessage := { ;
   {   0, 'Unknown Status '}, ;
   { 101, 'No User ID Passed to Server' }, ;
   { 102, 'Invalid Administrator User ID' }, ;
   { 103, 'Invalid Administrator Password' }, ;
   { 104, 'No Users' }, ;
   { 105, 'Invalid User ID' }, ;
   { 106, 'User Already Exists' }, ;
   { 107, 'No User ID or Password Passed to Server' }, ;
   { 108, 'Invalid Password' }, ;
   { 109, 'Invalid User ID ( TO-USER )' }, ;
   { 110, 'File/Path not found'}, ;
   { 111, 'Not Authorized'}, ;
   { 112, 'Wait-For-Message Timeout'}, ;
   { 150, 'Database Failure' }, ;
   { 200, 'User Data Enclosed' }, ;
   { 201, 'User Successfully Deleted' }, ;
   { 202, 'User Successfully Added' }, ;
   { 203, 'No Messages' }, ;
   { 204, 'Message Enclosed' } , ;
   { 205, 'User Data Successfully Updated' }, ;
   { 206, 'User Info Enclosed' }, ;
   { 207, 'Message Sent' }, ;
   { 208, 'File Copied' }, ;
   { 209, 'File Deleted' }, ;
   { 210, 'New User Added'}, ;
   { 211, 'User Status Updated'}, ;
   { 212, 'Group List Enclosed'}, ;
   { 213, 'Date and Time Enclosed'}, ;
   { 214, 'Temp File Name Enclosed'}, ;
   { 300, 'Activity for Specified User'}, ;
   { 301, 'User has been Logged ON' }, ;
   { 302, 'User has been Logged OFF'}, ;
   { 303, 'Force user Log OFF'}, ;
   { 304, 'Force user SHUT DOWN'}, ;
   { 400, 'User Status has Changed'} }

nFound := AScan( aMessage, {|a|a[1]==nStatus})

IF nFound > 0
  RETURN '(' + Alltrim(Str(nStatus)) + ') ' + aMessage[nFound,2]
ENDIF

RETURN '(' + Alltrim(Str(nStatus)) + ') ' + 'Unknown Status'

* --------------

STATIC FUNCTION _ValidateUser( cUserID, cPassword, lCheckPassword )

IF Empty(cUserId)
  RETURN 101 // empty user id
ENDIF
IF !OpenUserFile()
  RETURN 150 // database failure
ENDIF

cUserID := Alltrim(Upper(cUserID))

IF cUserID == 'ADMIN' .AND. cPassword == 'testing' .AND. lCheckPassword
  RETURN -1
ENDIF

IF !Empty(cUserID)
  IMUSERS->(dbSeek(Upper(cUserID)))
ENDIF

IF !IMUSERS->(Found())
  RETURN 105 // invalid user id
ENDIF

IF !lCheckPassword
  RETURN 0
ENDIF

IF !(Alltrim(cPassword) == Alltrim(IMUSERS->password))
  RETURN 108 // invalid password
ENDIF

UpdateUserInfo( cUserId )

RETURN 0

* --------------

STATIC FUNCTION UpdateUserInfo( cUserId )

LOCAL nFound

cUserId := Upper(Alltrim(cUserId))

nFound := AScan( saUsers, {|a|a[1]==cUserID})
IF nFound == 0
  AAdd(saUsers,{ cUserID, Date(), Seconds(), 0, .t. })
ELSE
  saUsers[nFound,2] := Date()
  saUsers[nFound,3] := Seconds()
ENDIF

RETURN nil

* --------------

STATIC FUNCTION UserFlag( cUserId, nStatus, lVisible, cExclude )

LOCAL nFound, i, nOldStatus := 0

cUserId := Upper(Alltrim(cUserId))

IF cUserId == '$$$' // set all users flags
  IF Valtype(nStatus) == 'N'
    FOR i := 1 TO Len(saUsers)
      IF !Empty(cExclude) .AND. cExclude == saUsers[i,1]
        LOOP
      ELSEIF nStatus == 303 // Forced log off
        saUsers[i,4] := nStatus
      ELSEIF !(saUsers[i,4] == 204)
        saUsers[i,4] := nStatus
      ENDIF
    NEXT
  ENDIF
  IF Valtype(lVisible) == 'L'
    FOR i := 1 TO Len(saUsers)
      saUsers[i,5] := lVisible
    NEXT
  ENDIF
  RETURN nStatus
ENDIF

nFound := AScan( saUsers, {|a|a[1]==cUserID})
IF nFound == 0
  DEFAULT nStatus := 0
  AAdd(saUsers,{ cUserID, CtoD(''), 0, nStatus, lVisible })
  nFound := Len(saUsers)
ENDIF
nOldStatus := saUsers[nFound,4]
IF Valtype(nStatus) = 'N'
  saUsers[nFound,4] := nStatus
ENDIF
IF Valtype(lVisible) = 'L'
  saUsers[nFound,5] := lVisible
ENDIF

RETURN nOldStatus

* --------------

STATIC FUNCTION IsUserActive( cUserId )

LOCAL lActive := .f., i, oClient

cUserId := Upper(Alltrim(cUserId))

FOR i := 1 TO Len(saClientList)
  oClient := saClientList[i]:requestClient
  IF Valtype(oClient)=='O' .AND. oClient:getVar("USERID") == cUserId
    lActive := .t.
    EXIT
  ENDIF
NEXT

RETURN lActive

* --------------

STATIC FUNCTION IsUserVisible( cUserId )

LOCAL nFound, lVisible

cUserId := Upper(Alltrim(cUserId))

nFound := AScan( saUsers, {|a|Upper(a[1])==Upper(cUserID)})

IF nFound == 0
  RETURN .T.
ENDIF

lVisible := saUsers[nFound,5]
DEFAULT lVisible := .t.

RETURN lVisible

* ----------

// This function returns a different file name with each call,

STATIC FUNCTION TempFileName( cPath, cExt )

LOCAL nSeconds := Seconds(), cFileName
STATIC snCount := 0

DEFAULT cExt := '.IM'

DO WHILE Seconds() - nSeconds < 100
  IF ++ snCount == 100
    snCount := 0
  ENDIF
  cFileName   := PadL( snCount, 2, "_" )    + ;
                 StrTran( Time(), ":", "" ) + ".$$$" + cExt

  IF !File(cPath+cFileName)
    EXIT
  ENDIF
ENDDO

RETURN cFileName

* ----------------

STATIC FUNCTION _DbSel( cAlias )

LOCAL i

cAlias := IIF(Valtype(cAlias)='C',UPPER(cAlias),'')
IF cAlias=='0'
  FOR i := 253 TO 1 STEP -1
    IF Empty(Alias(i))
      SELE (i)
      RETURN .t.
    ENDIF
  NEXT
  RETURN .f.
ENDIF
cAlias := IIF(Valtype(cAlias)='C',cAlias,'')
IF SELECT(cAlias)>0 .AND. !(':'$cAlias)
  SELE SELECT(cAlias)
  RETURN .t.
ENDIF
RETURN .f.

* -------------------

STATIC FUNCTION LogActivity ( oBrowse, xText, lSSL, lRefresh )

LOCAL i, cText, nBufferLen, cFileName, cMemo

DEFAULT lSSL := .F.

IF Valtype(xText) == 'A'
  cText := ''
  FOR i := 1 TO Len(xText)
    cText += DC_XtoC(xText[i]) + CRLF
  NEXT
ELSE
  cText := xText
ENDIF

IF !_DbSel('IMSERVER')
  OpenLogFile()
ENDIF

IF IMSERVER->(DC_AddRec(5))
  REPL IMSERVER->date WITH Date(), ;
       IMSERVER->time WITH Time(), ;
       IMSERVER->ssl WITH lSSL, ;
       IMSERVER->info WITH cText
  IMSERVER->(dbUnlock())
ENDIF

IF Valtype(oBrowse) = 'O' .AND. slRefreshBrowse
  oBrowse:goBottom()
  oBrowse:forceStable()
  oBrowse:refreshAll()
ENDIF

RETURN nil

* ------------------

STATIC FUNCTION OpenLogFile()

LOCAL aStru, cFileName := 'IMSERVER.DBF'

IF !File(cFileName)
  aStru := { ;
    { 'DATE', 'D', 8, 0 }, ;
    { 'TIME', 'C', 8, 0 }, ;
    { 'SSL',  'L', 1, 0 }, ;
    { 'INFO', 'C', 100, 0 } }
  dbCreate( cFileName, aStru, 'DBFCDX' )
ENDIF

USE (cFileName)

RETURN .t.

* -----------------

STATIC FUNCTION _LoadRdds()

LOCAL i, aDbeList := DbeList(), cDbeList := '', cDbe := dbeSetDefault()

IF Valtype(aDbeList) = 'A'
  FOR i := 1 TO Len(aDbeList)
    cDbeList += aDbeList[i,1] + ','
  NEXT
ENDIF

aDbeList := cDbeList

IF !('CDXDBE'$aDbeList) .AND. !DbeLoad( "CDXDBE",.T.)
  DC_WinAlert( "Database-Engine CDXDBE not loaded" )
ENDIF

IF !('DBFCDX'$aDbeList) .AND. !DbeBuild( "DBFCDX", "DBFDBE", "CDXDBE" )
  DC_WinAlert( "DBFCDX Database-Engine, Could not build engine" )
ENDIF

IF !('FOXDBE'$aDbeList) .AND. !DbeLoad( "FOXDBE",.T.)
  DC_WinAlert( "Database-Engine FOXDBE not loaded" )
ENDIF

IF !('FOXCDX'$aDbeList) .AND. !DbeBuild( "FOXCDX", "FOXDBE", "CDXDBE" )
  DC_WinAlert( "FOXCDX Database-Engine, Could not build engine" )
ENDIF

DbeSetDefault('FOXCDX')

RETURN .t.

* --------------

STATIC FUNCTION _SetUpData()

LOCAL cFolder := CurDir(), j, cParentFolder := ''

DO WHILE ( j := AT("\", cFolder) ) > 0
  cParentFolder += LEFT(cFolder, j)
  cFolder := SUBSTR(cFolder, j+1)
ENDDO
PUBLIC cDataPath := DC_Path(AppName(.t.))

SET DEFAULT TO (M->cDataPath)
SET EXACT OFF
SET EXCLUSIVE OFF
SET COLLATION TO AMERICAN
dbeSetDefault('FOXCDX')

RETURN nil

* ----------------

/*

FUNCTION ServiceRequest( oThread, cRequest )

IF (upper(cRequest) == 'SHUTDOWN')
  RETURN '***QUIT***'
ENDIF

RETURN 'Request not recognized'

* -----------

FUNCTION SID( oThread, nMode )

DEFAULT nMode := 1

IF nMode == 0 // Href
  RETURN 'SID=' + oThread:getSessionHandle()
ELSEIF nMode == 1 // Href
  RETURN '&SID=' + oThread:getSessionHandle()
ELSEIF nMode == 2 // aInclude
  RETURN { 'SID', oThread:getSessionHandle() }
ELSEIF nMode == 3 // input tag
  RETURN '<input type="hidden" name="SID" value=' + ;
          oThread:getSessionHandle() + '>'
ENDIF

RETURN ''

* ----------

Function FilterRequest(oClient)

 LOCAL cPath := Upper(oClient:httpRequest:path()), oThread := ThreadObject(), ;
       cResponse, oSession := oThread:getCargo("SESSION"), lStatus := .t.

if "WINNT" $ cPath .OR. 'WINDOWS' $ cPath
  // don't allow access to system dir
  oClient:HTTPResponse:ContentType("text/plain")
  oClient:HTTPResponse:Content := "The POLICE have been dispatched!"
  lStatus := FALSE
elseif '.PRG' $ cPath  .OR. '.EXE' $ cPath .OR. '.DLL' $ cPath
  oClient:HTTPResponse:ContentType("text/plain")
  oClient:HTTPResponse:Content := "No way.  Nice Try!"
  lStatus := FALSE
elseif '.ASP' $ cPath
  cPath := StrTran(cPath,'.ASP','.HTM')
  oClient:HTTPRequest:Path(cPath)
endif

Return lStatus

*/

* --------------

STATIC FUNCTION OpenUserFile( lTestStru )

LOCAL aStru, cFileName := 'IMUSERS.DBF', cIndexName

IF DC_DbSel('IMUSERS')
  IMUSERS->(OrdSetFocus('USER_ID'))
  RETURN .t.
ENDIF

DEFAULT lTestStru := .F.
IF lTestStru
  aStru := { ;
           { 'USER_ID   ','C',   50, 0 },;
           { 'PASSWORD  ','C',   20, 0 },;
           { 'USER_EMAIL','C',   50, 0 },;
           { 'USER_PHONE','C',   25, 0 },;
           { 'USER_LAST ','C',   30, 0 },;
           { 'USER_FIRST','C',   30, 0 },;
           { 'USER_ALIAS','C',   20, 0 },;
           { 'ACTIVE    ','L',    1, 0 },;
           { 'STATUS    ','C',   25, 0 },;
           { 'MODE      ','N',    3, 0 },;
           { 'USER_GROUP','C',  100, 0 },;
           { 'PREFERENCE','C',  100, 0 },;
           { 'LANGUAGE  ','C',    2, 0 },;
           { 'CLIENT_VER','C',   20, 0 }}
ENDIF

IF !File(cFileName)
  dbCreate( cFileName, aStru, 'DBFCDX' )
ENDIF

USE (cFileName) NEW
cIndexName := Strtran(cFileName,'.DBF','.CDX')

IF lTestStru .AND. !DC_IsStru( aStru )
  USE (cFileName) EXCL
  DC_StruUpdate( aStru )
  FErase(cIndexName)
ENDIF

IF !File(cIndexName)
  USE (cFileName) EXCLUSIVE
  INDEX ON USER_ID TAG 'USER_ID' TO (cIndexName)
ELSE
  SET INDEX TO (cIndexName)
ENDIF

IMUSERS->(OrdSetFocus('USER_ID'))

IF IMUSERS->(RecCount()) == 0
  IMUSERS->(DC_AddRec())
  IMUSERS->user_id := 'Administrator'
  IMUSERS->password := 'SilverOak1999'
  IMUSERS->user_first := 'System'
  IMUSERS->user_last := 'Administrator'
  IMUSERS->user_email := 'rogerdonnay@donnay-software.com'
  IMUSERS->user_alias := 'Admin'
  IMUSERS->user_group := 'ADMINISTRATORS'
  IMUSERS->active := .T.
  IMUSERS->(dbRUnlock())
ENDIF

RETURN .t.

* --------------

STATIC FUNCTION OpenGroupFile( lTestStru )

LOCAL aStru, cFileName := 'IMGROUPS.DBF', cIndexName

DEFAULT lTestStru := .F.
IF lTestStru
  aStru := { ;
           { 'GROUP_NAME','C',   20, 0 },;
           { 'GROUP_DESC','C',   50, 0 },;
           { 'GROUP_WAVE','C',   50, 0 } }
ENDIF

IF !File(cFileName)
  dbCreate( cFileName, aStru, 'DBFCDX' )
ENDIF

USE (cFileName) NEW
cIndexName := Strtran(cFileName,'.DBF','.CDX')

IF lTestStru .AND. !DC_IsStru( aStru )
  USE (cFileName) EXCL
  DC_StruUpdate( aStru )
  FErase(cIndexName)
ENDIF

IF !File(cIndexName)
  USE (cFileName) EXCLUSIVE
  INDEX ON GROUP_NAME TAG 'GROUP_NAME' TO (cIndexName)
ELSE
  SET INDEX TO (cIndexName)
ENDIF

IMGROUPS->(OrdSetFocus('GROUP_NAME'))

IF IMGROUPS->(RecCount()) == 0
  IMGROUPS->(DC_AddRec())
  IMGROUPS->group_name := 'ADMIN'
  IMGROUPS->group_desc := 'Administrators'
  IMGROUPS->(dbRUnlock())
ENDIF

RETURN .t.

* --------------

STATIC FUNCTION OpenMessageFile( lTestStru )

LOCAL aStru, cFileName := 'IMMSG.DBF', cIndexName

DEFAULT lTestStru := .f.

IF !File(cFileName) .OR. lTestStru
  aStru := { ;
           { 'TO_USER   ','C',   50, 0 },;
           { 'FROM_USER ','C',   50, 0 },;
           { 'DATE      ','D',    8, 0 },;
           { 'TIME      ','C',    8, 0 },;
           { 'MESSAGE   ','C', 1000, 0 },;
           { 'FONT      ','C',   30, 0 },;
           { 'FGCOLOR   ','N',   10, 0 },;
           { 'BGCOLOR   ','N',   10, 0 },;
           { 'FILE_NAME ','C',   50, 0 },;
           { 'FILE_TEMP ','C',   50, 0 },;
           { 'FILE_SIZE ','N',   10, 0 },;
           { 'WAVE      ','C',   50, 0 },;
           { 'CONV_NMBR ','N',   10, 0 },;
           { 'MSG_READ  ','L',    1, 0 },;
           { 'USER_LIST ','C',  100, 0 },;
           { 'CLONE     ','L',    1, 0 },;
           { 'MSG_TYPE  ','N',    1, 0 } }
  IF !File(cFileName)
    dbCreate( cFileName, aStru, 'DBFCDX' )
  ENDIF
ENDIF

USE (cFileName) NEW
cIndexName := Strtran(cFileName,'.DBF','.CDX')

IF lTestStru .AND. !DC_IsStru( aStru )
  USE (cFileName) EXCL
  DC_StruUpdate( aStru )
  FErase(cIndexName)
ENDIF

IF !File(cIndexName)
  USE (cFileName) EXCLUSIVE
  INDEX ON TO_USER + DtoS(DATE) + TIME TAG 'TO_USER' TO (cIndexName)
  INDEX ON FROM_USER + DtoS(DATE) + TIME TAG 'FROM_USER' TO (cIndexName)
  INDEX ON Str(CONV_NMBR,10) + DtoS(DATE) + TIME TAG 'CONV_NMBR' TO (cIndexName)
  INDEX ON DtoS(DATE) TAG 'DATE' TO (cIndexName)
  INDEX ON TO_USER + DtoS(DATE) + TIME TAG 'UNREAD_MSG' TO (cIndexName) ;
     FOR !MSG_READ
  INDEX ON Str(CONV_NMBR,10) + DtoS(DATE) + TIME TAG 'CONVD_UNIQ' TO (cIndexName) UNIQUE
  INDEX ON Str(CONV_NMBR,10) TAG 'CONV_UNIQ' UNIQUE
ELSE
  SET INDEX TO (cIndexName)
ENDIF

RETURN .t.

* --------------

STATIC FUNCTION _SendMessage( aToUserList, cFromUser, oRequest, oResponse, cWaveFile, ;
                              nConversationNum, lActiveOnly, nMessageType )

LOCAL nStatus, cMessage, aActive, i, cToUser, aTokens, cUserList, dDate, cTime

cMessage := oRequest:GetVar("Message")
IF !OpenMessageFile()
  RETURN 150
ENDIF

DEFAULT nMessageType := 0, ; // IM = 0 or 1, Alert = 2
        lActiveOnly := .t.

IF Empty(nConversationNum)
  IMMSG->(OrdSetFocus('CONV_NMBR'))
  IMMSG->(dbGoBottom())
  nConversationNum := IMMSG->conv_nmbr + 1
ENDIF

IMMSG->(OrdSetFocus('TO_USER'))
aActive := Array(Len(aToUserList))
AFill(aActive,.f.)

IF !Empty(cMessage) .AND. Left(cMessage,10) == '<<invite>>'
  aTokens := DC_TokenArray(cMessage,',')
ENDIF

cUserList := cFromUser + ','
FOR i := 1 TO Len(aToUserList)
  cUserList += aToUserList[i]
  IF i < Len(aToUserList)
    cUserList += ','
  ENDIF
NEXT

dDate := Date()
cTime := Time()

FOR i := 1 TO Len(aToUserList)

  Sleep(1)
  cToUser := aToUserList[i]

  aActive[i] := IsUserActive( cToUser )
  IF !aActive[i] .AND. lActiveOnly
    LOOP
  ENDIF

  IMMSG->(OrdSetFocus('TO_USER'))
  IMMSG->(dbSeek(Space(USERID_LEN)))
  IF IMMSG->(Eof()) .AND. snDays > 0
    IMMSG->(OrdSetFocus('DATE'))
    IMMSG->(dbSeek(DtoS(Date()+snDays),.t.))
  ENDIF
  IF IMMSG->(Eof())
    IF !IMMSG->(DC_AddRec(5))
      RETURN 150
    ENDIF
  ELSE
    IF !IMMSG->(DC_RecLock(5))
      RETURN 150
    ENDIF
  ENDIF

  IMMSG->msg_type := nMessageType
  IMMSG->message := DC_XtoC(cMessage)
  IMMSG->to_user := cToUser
  IMMSG->from_user := cFromUser
  IMMSG->date := dDate
  IMMSG->time := cTime
  IMMSG->fgColor := Val(DC_XtoC(oRequest:getVar('FGColor')))
  IMMSG->bgColor := Val(DC_XtoC(oRequest:getVar('BGColor')))
  IMMSG->font := DC_XtoC(oRequest:getVar('Font'))
  IMMSG->file_name := DC_XtoC(oRequest:getVar('FileName'))
  IMMSG->file_temp := DC_XtoC(oRequest:getVar('TempFileName'))
  IMMSG->conv_nmbr := nConversationNum
  IMMSG->msg_read := .f.
  IMMSG->user_list := cUserList
  IF !Empty(cWaveFile)
    IMMSG->wave := cWaveFile
  ELSE
    IMMSG->wave := DC_XtoC(oRequest:getVar('WaveFile'))
  ENDIF

  IMMSG->(dbRUnlock())
  UserFlag( cToUser, 207 )

  IF !Empty(aTokens) .AND. Upper(Alltrim(aTokens[3])) == Upper(Alltrim(cToUser))
    AddConferenceMessages( cToUser, nConversationNum )
  ENDIF

NEXT
oResponse:setVar('ActiveArray',aActive)
oResponse:setVar('ConversationNum',nConversationNum)
oResponse:setVar('ServerTime',cTime)
oResponse:setVar('ServerDate',dDate)

RETURN 207

* ---------------

STATIC FUNCTION AddConferenceMessages( cToUser, nConversationNum )

LOCAL aRecords[0], aData, i

IMMSG->(OrdSetFocus('CONVD_UNIQ'))
IMMSG->(dbSeek(Str(nConversationNum,10)))

DO WHILE IMMSG->conv_nmbr == nConversationNum .AND. !IMMSG->(Eof())
  aData := IMMSG->(Scatter())
  aData[1] := cToUser
  IF Left(IMMSG->message,2) # '<<' .AND. !IMMSG->clone
    AAdd(aRecords,aData)
  ENDIF
  IMMSG->(dbSkip())
ENDDO

FOR i := 1 TO Len(aRecords)
  aData := aRecords[i]
  IF IMMSG->(DC_AddRec(5))
    IMMSG->(Gather(aData))
    IMMSG->msg_read := .f.
    IMMSG->date := Date()
    IMMSG->time := Time()
    IMMSG->clone := .t.
    IMMSG->(dbRUnlock())
  ENDIF
NEXT

RETURN nil

* ---------------

DLLFUNCTION compress(@cOut, @nOut, @cIn, nIn) USING CDECL FROM ZLIB1.DLL
DLLFUNCTION compress2(@cOut, @nOut, @cIn, nIn, nLevel) USING CDECL FROM ZLIB1.DLL
DLLFUNCTION uncompress(@cOut, @nOut, @cIn, nIn) USING CDECL FROM ZLIB1.DLL
DLLFUNCTION CreateDirectoryA( a, b ) USING STDCALL FROM KERNEL32.DLL

* ---------------

FUNCTION UpdateUserFile( cIniFile )

LOCAL cUserId, aData, aRef, cUserDatabase, cUserDbe

aRef := { 'Database', ;
          'Dbe', ;
          'Field_UserId', ;
          'Field_Password', ;
          'Field_Email', ;
          'Field_LastName', ;
          'Field_FirstName', ;
          'Field_Alias', ;
          'Field_UserPhone',;
          'Password_Literal' }

aData := { Space(100), ;
           'FOXCDX     ', ;
           Space(50), ;
           Space(50), ;
           Space(50), ;
           Space(50), ;
           Space(50), ;
           Space(50), ;
           Space(50), ;
           Space(30) }

DC_IniLoad( cIniFile, 'USERDATA', aData, aRef )

cUserDatabase     := Alltrim(aData[1])
cUserDbe          := Alltrim(aData[2])
PRIVATE cField_UserId     := Alltrim(aData[3])
PRIVATE cField_Password   := Alltrim(aData[4])
PRIVATE cField_Email      := Alltrim(aData[5])
PRIVATE cField_Lastname   := Alltrim(aData[6])
PRIVATE cField_Firstname  := Alltrim(aData[7])
PRIVATE cField_Alias      := Alltrim(aData[8])
PRIVATE cField_UserPhone  := Alltrim(aData[9])
PRIVATE cPassword         := Alltrim(aData[10])

IF Empty(cUserDatabase)
  RETURN .f.
ENDIF

USE (cUserDatabase) VIA (cUserDbe) ALIAS 'USR'

OpenUserFile(.t.,.t.)
IMUSERS->(dbGoTop())
IMUSERS->(OrdSetFocus(0))
DO WHILE !IMUSERS->(Eof())
  IF !(IMUSERS->user_id = 'Administrator') .AND. IMUSERS->(DC_RecLock(5))
    IMUSERS->active := .f.
  ENDIF
  IMUSERS->(dbSkip())
ENDDO
IMUSERS->(OrdSetFocus('USER_ID'))
USR->(dbGoTop())
SET DELETED OFF

DO WHILE !USR->(Eof())
  cUserId := Pad(Upper(USR->(&(M->cField_UserID))),50)
  IMUSERS->(dbSeek(cUserId))
  IF IMUSERS->(Found())
    IF IMUSERS->(DC_RecLock(5))
      IF !Empty(M->cField_Email)
        IMUSERS->user_email := USR->(&(M->cField_Email))
      ENDIF
      IF !Empty(M->cField_LastName)
        IMUSERS->user_last := USR->(&(M->cField_LastName))
      ENDIF
      IF !Empty(M->cField_FirstName)
        IMUSERS->user_first := USR->(&(M->cField_FirstName))
      ENDIF
      IF Empty(IMUSERS->user_alias)
        IMUSERS->user_alias := USR->(&(M->cField_UserId))
      ENDIF
      IF !Empty(M->cField_Password)
        IMUSERS->password := USR->(&(M->cField_Password))
      ELSEIF !Empty(M->cPassword)
        IMUSERS->password := M->cPassword
      ENDIF
      IF !Empty(M->cField_UserPhone)
        IMUSERS->user_phone := USR->(&(M->cField_UserPhone))
      ENDIF
      IMUSERS->active := .t.
    ENDIF
  ELSE
    IF IMUSERS->(DC_AddRec(5))
      IMUSERS->user_id := USR->(&(M->cField_UserID))
      IF !Empty(M->cField_Password)
        IMUSERS->password := USR->(&(M->cField_Password))
      ELSEIF !Empty(M->cPassword)
        IMUSERS->password := M->cPassword
      ENDIF
      IF !Empty(M->cField_Email)
        IMUSERS->user_email := USR->(&(M->cField_Email))
      ENDIF
      IF !Empty(M->cField_Lastname)
        IMUSERS->user_last := USR->(&(M->cField_LastName))
      ENDIF
      IF !Empty(M->cField_FirstName)
        IMUSERS->user_first := USR->(&(M->cField_FirstName))
      ENDIF
      IF !Empty(M->cField_Alias)
        IMUSERS->user_alias := USR->(&(M->cField_Alias))
      ENDIF
      IF !Empty(M->cField_UserPhone)
        IMUSERS->user_phone := USR->(&(M->cField_UserPhone))
      ENDIF
      IMUSERS->active := .t.
      IMUSERS->(dbUnlock())
    ENDIF
  ENDIF
  USR->(dbSkip())
ENDDO

SELE IMUSERS
DELETE FOR !IMUSERS->active
IMUSERS->(dbPack())
IMUSERS->(dbReindex())
IMUSERS->(dbCloseArea())
USR->(dbCloseArea())

RETURN nil

* ------------

FUNCTION FirstName( cName )

LOCAL aTokens := DC_TokenArray( cName, ' ' )
IF Len(aTokens) < 1
  RETURN ''
ENDIF

RETURN aTokens[1]

* ------------

FUNCTION LastName( cName )

LOCAL aTokens := DC_TokenArray( cName, ' ' )

IF Len(aTokens) < 2
  RETURN ''
ENDIF

RETURN aTokens[2]

* --------------

STATIC FUNCTION SendMailThread(lRunning,aMainGetList)

LOCAL cSmtpServer, cDbe, cUserId, cPassword, cSender, ;
      cRecipient, cDatabaseName, cErrorText, cMailTo, ;
      cMailFrom, cReplyTo, cSubject, cBody, lError, ;
      cAttachment

LOCAL bErrorBlock := ErrorBlock({|e|TrapError(e)})

LOCAL aData := { Pad('donnay-software.com',100), ;
                 'FOXCDX   ', ;
                 Space(50), ;
                 Space(50), ;
                 Pad('rogerdonnay@donnay-software.com',50), ;
                 Pad('rogerdonnay@donnay-software.com',50), ;
                 Pad(DC_Path(AppName(.t.))+'SENDMAIL.DBF',100)}

LOCAL aRef := { 'SMTPserver', ;
                'DatabaseDbe', ;
                'UserId', ;
                'Password', ;
                'Sender', ;
                'Recipient', ;
                'DatabaseName' }

LOCAL cIniFile := DC_Path(AppName(.t.)) + 'IMSERVER.INI'
DC_IniLoad( cIniFile, 'EMAIL', aData, aRef )

lRunning := .f.

cSmtpServer := Alltrim(aData[1])
cDbe := Alltrim(aData[2])
cUserId := Alltrim(aData[3])
cPassword := Alltrim(aData[4])
cSender := Alltrim(aData[5])
cRecipient := Alltrim(aData[6])
cDatabaseName := Alltrim(aData[7])

BEGIN SEQUENCE

lError := .t.
OpenSendMailFile( cDatabaseName, cDbe, .t. )

lError := .f.
END SEQUENCE

ErrorBlock(bErrorBlock)

IF lError
  DCMSGBOX 'Error opening ' + cDatabaseName, ;
           'Closing SendMail Thread!' TIMEOUT 2
  RETURN .f.
ENDIF

lRunning := .t.
DO WHILE lRunning

  IF slIsPaused
    Sleep(100)
    LOOP
  ENDIF

  SENDMAIL->(dbGoTop())
  IF !SENDMAIL->(Eof())

    cMailTo := Alltrim(SENDMAIL->mail_to)
    cMailFrom := Alltrim(SENDMAIL->mail_from)
    cReplyTo := Alltrim(SENDMAIL->reply_to)
    cSubject := Alltrim(SENDMAIL->subject)
    cBody := Alltrim(SENDMAIL->body) + ' (' + Alltrim(SENDMAIL->user_id) + ')'
    cAttachment := Alltrim(SENDMAIL->attachment)

    IF Empty(cMailTo)
      cMailTo := cRecipient
    ENDIF
    IF Empty(cMailFrom)
      cMailFrom := cSender
    ENDIF
    IF Empty(cReplyTo)
      cReplyTo := cMailFrom
    ENDIF

    cErrorText := ''
    IF SendMail( cSMTPServer, cMailFrom, cMailTo, cSubject, ;
                 cReplyTo, cAttachment, cBody, @cErrorText, ;
                 cUserId, cPassword )
      IF SENDMAIL->(DC_RecLock())
        SENDMAIL->date_sent := Date()
        SENDMAIL->time_sent := Time()
        SENDMAIL->mail_to := cMailTo
        SENDMAIL->mail_from := cMailFrom
        SENDMAIL->reply_to := cReplyTo
        SENDMAIL->(dbRUnlock())
      ENDIF
    ELSE
      IF SENDMAIL->(DC_RecLock())
        SENDMAIL->error_code := 1
        SENDMAIL->error_text := cErrorText
        SENDMAIL->(dbRUnlock())
      ENDIF
    ENDIF
  ENDIF
  Sleep(1000)
  DC_GetRefresh(aMainGetList,'USERS')
ENDDO

RETURN nil

* --------------

STATIC FUNCTION TrapError()
BREAK
RETURN .t.

* --------------

STATIC FUNCTION SendMail( cSMTPServer, cSender, cRecipient, cSubject, ;
                          cReplyTo, cAttachment, cText, cErrorText, ;
                          cUserId, cPassword )

LOCAL oMail, oSender, oRecipient, oSmtp, lError := .t., cPrefix

DEFAULT cSMTPServer := "donnay-software.com", ;
        cSender := 'rogerdonnay@donnay-software.com', ;
        cRecipient := 'rogerdonnay@donnay-software.com'

IF Empty(cSMTPServer)
  cErrortext := 'No SMTP Server Address'
ELSEIF Empty(cSender)
  cErrorText := 'No Sender Address'
ELSEIF Empty(cRecipient)
  cErrorText := 'No Recipient Address'
ELSEIF Empty(cSubject)
  cErrortext := 'No Subject'
ELSE
  lError := .f.
ENDIF
IF lError
  dcbdebug cErrorText
  RETURN .f.
ENDIF

oMail      := MIMEMessage():new()
oSender    := MailAddress():new( Alltrim(cSender) )
oRecipient := MailAddress():new( Alltrim(cRecipient) )

// Assemble the e-mail
oMail:setFrom( oSender     )
oMail:setSubject( Alltrim(cSubject) )
oMail:setMessage( cText )
oMail:addRecipient( oRecipient  )
IF !Empty(cReplyTo)
  oMail:addHeader( "Reply-To", Alltrim(cReplyTo) )
ENDIF
oMail:setContentType('text')

IF !Empty(cAttachment)
  oMail:attachFile( cAttachment  )
ENDIF

oSmtp := SMTPClient():new( Alltrim(cSMTPServer) )
IF oSmtp:connect( cUserId, cPassword )
  oSmtp:send( oMail )
  oSmtp:disconnect()
  IF !Empty(cAttachment) .AND. 'F:\ENTERPRISE\' $ Upper(cAttachment)
    FErase(cAttachment)
  ENDIF
  lError := .f.
ELSE
  cErrorText := "Unable to connect to mail server"
ENDIF

IF !Empty(cErrorText)
  lError := .t.
  DCBDEBUG cErrorText
ENDIF

RETURN !lError

* ------------

STATIC FUNCTION OpenSendMailFile( cFileName, cDbe, lTestStru )

LOCAL aStru, cIndexName

DEFAULT lTestStru := .f., ;
        cFileName := 'SENDMAIL.DBF', ;
        cDbe := 'FOXCDX'

IF !File(cFileName) .OR. lTestStru
  aStru := { ;
           { 'USER_ID',   'C',   20, 0 }, ;
           { 'USER_NAME', 'C',   30, 0 }, ;
           { 'STATION',   'C',    3, 0 }, ;
           { 'MAIL_TO',   'C',  100, 0 }, ;
           { 'MAIL_FROM', 'C',  100, 0 }, ;
           { 'REPLY_TO',  'C',  100, 0 }, ;
           { 'SUBJECT',   'C',   50, 0 }, ;
           { 'ATTACHMENT','C',  250, 0 }, ;
           { 'BODY',      'M',   10, 0 }, ;
           { 'DATE',      'D',    8, 0 }, ;
           { 'TIME',      'C',    8, 0 }, ;
           { 'DATE_SENT', 'D',    8, 0 }, ;
           { 'TIME_SENT', 'C',    8, 0 }, ;
           { 'ERROR_CODE','N',    1, 0 }, ;
           { 'ERROR_TEXT','C',   50, 0 } }
  IF !File(cFileName)
    dbCreate( cFileName, aStru, cDbe )
  ENDIF
ENDIF

USE (cFileName) NEW ALIAS 'SENDMAIL'
cIndexName := Strtran(cFileName,'.DBF','.CDX')

IF lTestStru .AND. !DC_IsStru( aStru )
  USE (cFileName) EXCL
  DC_StruUpdate( aStru )
  FErase(cIndexName)
ENDIF

IF !File(cIndexName)
  USE (cFileName) EXCLUSIVE ALIAS 'SENDMAIL'
  INDEX ON mail_to TAG 'MAILTO' TO (cIndexName)
  INDEX ON mail_from TAG 'MAILFROM' TO (cIndexName)
  INDEX ON DtoS(date)+time TAG 'DATE' TO (cIndexName)
  INDEX ON user_id TAG 'UNSENT' FOR Empty(date_sent).AND.Empty(error_code)
ELSE
  SET INDEX TO (cIndexName)
ENDIF

RETURN .t.

* -----------

STATIC FUNCTION ShowUsers()

LOCAL i, oClient, nClients, aUsers[0], GetList[0], oBrowse, cUserId

nClients := Len(saClientList)

IF !OpenUserFile()
  RETURN .f.
ENDIF

FOR i := 1 TO nClients
  oClient := saClientList[i]:requestClient
  IF Valtype(oClient) == 'O'
    cUserID := Pad(oClient:getVar("USERID"),50)
    IMUSERS->(dbSeek(cUserId))
    IF AScan( aUsers, {|a|Upper(Trim(a[1])) == Upper(Trim(cUserID))}) == 0
      AAdd( aUsers, { Trim(cUserId), Trim(IMUSERS->user_first) + ' ' + Trim(IMUSERS->user_last) } )
    ENDIF
  ENDIF
NEXT

@ 0,0 DCBROWSE oBrowse DATA aUsers PRESENTATION DC_BrowPres();
      SIZE 20,20 FIT FONT '10.Courier New'

DCBROWSECOL ELEMENT 1 HEADER 'User ID' WIDTH 20 PARENT oBrowse
DCBROWSECOL ELEMENT 2 HEADER 'Name' WIDTH 25 PARENT oBrowse

DCREAD GUI FIT TITLE 'Logged On Users' MODAL

IMUSERS->(dbCloseArea())

RETURN .t.

